import functools

import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.delivery._data._data_provider_factory import make_provider
from tests.unit.conftest import StubSession


def test_contracts_version():
    # given
    expected = "1.0.155"

    # when
    testing_version = rdf.__version__

    # then
    assert testing_version == expected, testing_version


def test_definition_repr():
    # given
    expected_value = (
        "<refinitiv.data.content.ipa.financial_contracts.Definition object at"
    )
    definition = rdf.Definitions([])

    # when
    testing_value = repr(definition)

    # then
    assert expected_value in testing_value, testing_value


def test_definition_call():
    # given
    bond_definition = rdf.bond.Definition(
        issue_date="2002-02-28",
        end_date="2032-02-28",
        notional_ccy="USD",
        interest_payment_frequency="Annual",
        fixed_rate_percent=7,
        interest_calculation_method=rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL,
    )

    # when
    definition = rdf.Definitions([bond_definition])

    # then
    assert definition


def test_request_body():
    # given
    expected_body = {
        "fields": ["1", "2"],
        "pricingParameters": {"marketDataDate": "market_data_date"},
        "universe": [
            {
                "instrumentDefinition": {
                    "extended_params": "extended_params",
                    "legs": [{"paymentBusinessDays": "payment_business_days"}],
                    "startDate": "start_date",
                },
                "instrumentType": "Swap",
                "pricingParameters": {"marketDataDate": "market_data_date"},
            },
            {
                "instrumentDefinition": {
                    "endDate": "2032-02-28",
                    "extended_params": "extended_params",
                    "fixedRatePercent": 7,
                    "interestCalculationMethod": "Dcb_Actual_Actual",
                    "interestPaymentFrequency": "Annual",
                    "issueDate": "2002-02-28",
                    "notionalCcy": "USD",
                },
                "instrumentType": "Bond",
                "pricingParameters": {"tradeDate": "trade_date"},
            },
        ],
    }
    bond_definition = rdf.bond.Definition(
        issue_date="2002-02-28",
        end_date="2032-02-28",
        notional_ccy="USD",
        interest_payment_frequency="Annual",
        fixed_rate_percent=7,
        interest_calculation_method=rdf.bond.DayCountBasis.DCB_ACTUAL_ACTUAL,
        extended_params={"extended_params": "extended_params"},
        pricing_parameters=rdf.bond.PricingParameters(trade_date="trade_date"),
        fields=["1", "2"],
    )
    swap_definition = rdf.swap.Definition(
        start_date="start_date",
        fields=["1", "2"],
        legs=[rdf.swap.LegDefinition(payment_business_days="payment_business_days")],
        pricing_parameters=rdf.swap.PricingParameters(
            market_data_date="market_data_date"
        ),
        extended_params={"extended_params": "extended_params"},
    )
    definition = rdf.Definitions(
        universe=[swap_definition, bond_definition],
        fields=["1", "2"],
        pricing_parameters=rdf.swap.PricingParameters(
            market_data_date="market_data_date"
        ),
    )

    # when
    content_type = definition._kwargs["__content_type__"]
    provider = make_provider(content_type)
    request = provider.request.create(StubSession(), "url", **definition._kwargs)

    # then
    testing_body = request.json
    assert testing_body == expected_body, expected_body


@pytest.mark.parametrize(
    "definition",
    [
        functools.partial(rdf.Definitions, None),
        functools.partial(rdf.Definitions, ""),
        functools.partial(rdf.Definitions, "None"),
        functools.partial(rdf.Definitions, 100),
        functools.partial(rdf.Definitions, ["None"]),
        functools.partial(rdf.Definitions, {None}),
        functools.partial(rdf.Definitions, [{}]),
        functools.partial(rdf.Definitions, [100]),
        functools.partial(rdf.Definitions, rdf.Definitions),
        functools.partial(
            rdf.Definitions,
            [rdf.term_deposit.Definition(), "None", rdf.bond.Definition()],
        ),
    ],
)
def test_validate_universe_raise_type_error(definition):
    with pytest.raises(TypeError):
        definition()


def test_validate_universe_type_error_message():
    err_msg = (
        "Provided type for parameter 'universe' is invalid. "
        "Expected types: [bond.Definition, cap_floor.Definition, cds.Definition, cross.Definition, "
        "option.Definition, repo.Definition, swap.Definition, swaption.Definition, "
        "term_deposit.Definition]"
    )
    try:
        rdf.Definitions(None)
    except TypeError as e:
        assert str(e) == err_msg


@pytest.mark.parametrize(
    "definition",
    [
        functools.partial(rdf.Definitions, rdf.bond.Definition()),
        functools.partial(rdf.Definitions, rdf.cap_floor.Definition()),
        functools.partial(rdf.Definitions, rdf.cds.Definition()),
        functools.partial(rdf.Definitions, rdf.cross.Definition()),
        functools.partial(rdf.Definitions, rdf.option.Definition()),
        functools.partial(rdf.Definitions, rdf.repo.Definition()),
        functools.partial(rdf.Definitions, rdf.swap.Definition()),
        functools.partial(rdf.Definitions, rdf.swaption.Definition()),
        functools.partial(rdf.Definitions, rdf.term_deposit.Definition()),
        functools.partial(rdf.Definitions, [rdf.term_deposit.Definition()]),
        functools.partial(
            rdf.Definitions,
            [
                rdf.term_deposit.Definition(),
                rdf.term_deposit.Definition(),
                rdf.bond.Definition(),
            ],
        ),
        functools.partial(rdf.Definitions, set()),
        functools.partial(rdf.Definitions, dict()),
        functools.partial(rdf.Definitions, []),
    ],
)
def test_validate_universe_do_not_raise_type_error(definition):
    try:
        definition()
    except TypeError as e:
        assert False, str(e)
    else:
        assert True



from refinitiv.data.content.ipa._curves._models._curve import Curve
from refinitiv.data.content.ipa.curves import forward_curves, zc_curves


def forward_curve_definition_01():
    definition = forward_curves.Definition(
        forward_curve_definitions=[
            forward_curves.ForwardCurveDefinition(
                index_tenor="3M",
                forward_curve_tag="ForwardTag",
                forward_start_date="2021-02-01",
                forward_curve_tenors=["0D", "1D"],
                forward_start_tenor="some_start_tenor",
            )
        ],
        curve_definition=forward_curves.SwapZcCurveDefinition(
            currency="EUR",
            index_name="EURIBOR",
            discounting_tenor="OIS",
            name="EUR EURIBOR Swap ZC Curve",
        ),
        curve_parameters=forward_curves.SwapZcCurveParameters(
            calendar_adjustment=forward_curves.CalendarAdjustment.CALENDAR
        ),
        curve_tag="test_curve",
    )

    return definition


def forward_curve_definition_02():
    definition = forward_curves.Definition(
        forward_curve_definitions=[
            forward_curves.ForwardCurveDefinition(
                index_tenor="3M",
                forward_curve_tag="ForwardTag",
                forward_start_date="2021-02-01",
                forward_curve_tenors=["0D", "1D"],
                forward_start_tenor="some_start_tenor",
            )
        ],
        curve_definition=forward_curves.SwapZcCurveDefinition(
            currency="EUR",
            index_name="EURIBOR",
            discounting_tenor="OIS",
            name="EUR EURIBOR Swap ZC Curve",
        ),
        curve_parameters=forward_curves.SwapZcCurveParameters(),
        curve_tag="test_curve",
    )

    return definition


def forward_curve_definition_03():
    definition = forward_curves.Definition()

    return definition


def zc_curves_01():
    definition = zc_curves.Definition(
        curve_definition=zc_curves.ZcCurveDefinitions(
            currency="CHF",
            name="CHF LIBOR Swap ZC Curve",
            discounting_tenor="OIS",
        ),
        curve_parameters=zc_curves.ZcCurveParameters(
            valuation_date="2019-08-21", use_steps=True
        ),
    )

    return definition


def zc_curves_02():
    definition = zc_curves.Definition(
        constituents={},
        curve_parameters=zc_curves.ZcCurveParameters(
            valuation_date="2019-08-21",
            price_side="Mid",
            interpolation_mode=zc_curves.ZcInterpolationMode.CUBIC_DISCOUNT,
        ),
        curve_definition=zc_curves.ZcCurveDefinitions(
            currency="EUR",
            index_name="EURIBOR",
            source="Refinitiv",
            discounting_tenor="OIS",
            name="EUR EURIBOR Swap ZC Curve",
        ),
        curve_tag="TAG",
    )
    return definition


def check_curve(curve):
    assert isinstance(curve, Curve)
    assert curve.x.size > 0, f"Empty curve.x received"
    assert curve.y.size > 0, f"Empty curve.y received"


def check_curves(response):
    curves = response.data.curves
    for curve in curves:
        check_curve(curve)


def on_response(response, definition, session):
    print(response.data.df.to_string())


def check_curve_response(response, expected_currency):
    actual_currency = response.data.raw["data"][0]["curveDefinition"]["currency"]
    assert actual_currency == expected_currency, f"{actual_currency}"
    assert not response.data.df.empty



import json
import os
import uuid
from importlib import reload
from typing import TYPE_CHECKING

import pytest

import refinitiv.data as rd
import refinitiv.data.eikon as ek
from refinitiv.data import OpenState
from refinitiv.data import _configure as configure
from refinitiv.data.content import custom_instruments
from tests.integration.credentials import (
    config_for_beta_user,
    config_for_prod_user,
    Credentials,
)
from tests.integration.helpers import (
    write_prj_config,
    create_obj_with_value_by_path,
)

if TYPE_CHECKING:
    from refinitiv.data.delivery._stream.base_stream import StreamOpenMixin

CONFIG_PROD = "./tests/integration/refinitiv-data.custom.config-prod.json"
CONFIG_BETA = "./tests/integration/refinitiv-data.custom.config-beta.json"

PATH = "sessions.platform.default.base-url"
BASE_URL_BETA = "https://api.ppe.refinitiv.com"
ENV_NAME = os.environ.get("ENV_NAME")
RDP_PLATFORM = "rdp"
UDF_PLATFORM = "udf"

config_credentials = config_for_beta_user
conf = CONFIG_BETA

if ENV_NAME == "PROD":
    config_credentials = config_for_prod_user
    conf = CONFIG_PROD


def get_env_var(key):
    retval = os.environ.get(config_credentials[key])
    if retval is None:
        raise Exception(f"{config_credentials[key]} is not an env variable.")
    return retval


desktop_app_key = get_env_var("app_key")

edp = Credentials(get_env_var("username"), get_env_var("password"))
edp2 = Credentials(get_env_var("username_second"), get_env_var("password_second"))
edp3 = Credentials(get_env_var("username_third"), get_env_var("password_third"))

rdp = Credentials(get_env_var("rdp_username"), get_env_var("rdp_password"))
rdp2 = Credentials(
    get_env_var("rdp_username_second"), get_env_var("rdp_password_second")
)

eikon = Credentials(get_env_var("eikon_username"), get_env_var("eikon_password"))

PROJECT_CONFIG_PATH = os.path.join(os.getcwd(), configure._default_config_file_name)


def remove_config(path):
    while os.path.exists(path):
        try:
            os.remove(path)
            import time

            time.sleep(1)
        except (PermissionError, FileNotFoundError):
            pass


def remove_project_config():
    remove_config(PROJECT_CONFIG_PATH)


@pytest.fixture(scope="function")
def load_config():
    rd.load_config(conf)


@pytest.fixture()
def project_config_path():
    return PROJECT_CONFIG_PATH


@pytest.fixture(scope="function")
def write_project_config(project_config_path):
    def inner(arg):
        if isinstance(arg, dict):
            arg = json.dumps(arg)
        f = open(project_config_path, "w")
        f.write(arg)
        f.close()
        return project_config_path

    yield inner

    configure._dispose()
    remove_project_config()


@pytest.fixture(scope="function", autouse=False)
def set_env_base_url(write_project_config, credential_config):
    if ENV_NAME == "BETA":
        config = create_obj_with_value_by_path(PATH, BASE_URL_BETA)
        write_prj_config(config)
        reload(configure)


def _create_platform_session(credential=edp, scope="trapi"):
    return rd.session.platform.Definition(
        app_key=desktop_app_key,
        grant=rd.session.platform.GrantPassword(
            username=credential.username,
            password=credential.password,
            token_scope=scope,
        ),
    ).get_session()


def create_platform_session(credential=edp, scope="trapi"):
    return _create_platform_session(credential=credential, scope=scope)


def create_platform_session_with_rdp_creds():
    return _create_platform_session(credential=rdp)


def create_desktop_session():
    return rd.session.desktop.Definition(app_key=desktop_app_key).get_session()


@pytest.fixture(
    scope="function", params=[create_desktop_session], ids=["desktop_session"]
)
def open_desktop_session(set_env_base_url, request):
    create_session = request.param
    session = create_session()
    session.open()
    rd.session.set_default(session)
    yield session
    session.close()
    rd.session.set_default(None)


@pytest.fixture(scope="function")
def open_platform_session(set_env_base_url):
    session = create_platform_session()
    session.open()
    rd.session.set_default(session)
    yield session
    session.close()
    rd.session.set_default(None)


@pytest.fixture(scope="function")
def open_platform_session_with_rdp_creds(set_env_base_url):
    session = create_platform_session(credential=rdp)
    session.open()
    rd.session.set_default(session)
    yield session
    session.close()
    rd.session.set_default(None)


credential_sets = [
    (edp, rdp),
    (edp2, rdp2),
    (edp3, rdp),
]


@pytest.fixture()
def credential_config(request):
    if hasattr(request.config, "workerinput"):  # xdist
        node_id = int(request.config.workerinput["workerid"].replace("gw", ""))
        if (
            request.instance.pytestmark[0].args[0] == "tds"
        ):  # "tds" test suites access granted with cred #0
            node_id = 0

        global edp, rdp
        (edp, rdp) = credential_sets[node_id]


@pytest.fixture(
    scope="function",
    params=[create_platform_session, create_desktop_session],
    ids=["platform_session", "desktop_session"],
)
def open_session(request, set_env_base_url):
    create_session = request.param
    if request.instance is not None:
        if request.instance.pytestmark[0].args[0] == "news":
            set_underlying_param(request.param_index)
    session = create_session()
    session.open()
    rd.session.set_default(session)
    yield session
    session.close()
    rd.session.set_default(None)


@pytest.fixture(
    scope="function", params=[create_desktop_session], ids=["desktop_session"]
)
async def open_desktop_session_async(set_env_base_url, request):
    create_session = request.param
    session = create_session()
    await session.open_async()
    rd.session.set_default(session)
    yield session
    await session.close_async()
    rd.session.set_default(None)


@pytest.fixture(scope="function")
async def open_platform_session_async(set_env_base_url):
    session = create_platform_session()
    await session.open_async()
    rd.session.set_default(session)
    yield session
    await session.close_async()
    rd.session.set_default(None)


@pytest.fixture(scope="function")
async def open_platform_session_with_rdp_creds_async(set_env_base_url):
    session = create_platform_session(credential=rdp)
    await session.open_async()
    rd.session.set_default(session)
    yield session
    await session.close_async()
    rd.session.set_default(None)


@pytest.fixture(
    scope="function",
    params=[create_platform_session, create_desktop_session],
    ids=["platform_session", "desktop_session"],
)
async def open_session_async(request, set_env_base_url):
    create_session = request.param
    if request.instance is not None:
        if request.instance.pytestmark[0].args[0] == "news":
            set_underlying_param(request.param_index)
    session = create_session()
    await session.open_async()
    rd.session.set_default(session)
    yield session
    await session.close_async()
    rd.session.set_default(None)


def set_underlying_param(param):
    if param == 0:
        set_rdp_config("news")
    elif param == 1:
        set_udf_config("news")


def get_dict_values_as_list(dict_response):
    return [*dict_response.values()]


def get_dict_keys_as_list(dict_response):
    return [*dict_response.keys()]


def is_open(stream: "StreamOpenMixin") -> bool:
    return stream.open_state is OpenState.Opened


def set_udf_config(content_name):
    rd.get_config().set_param(
        param=f"apis.data.{content_name}.underlying-platform", value=UDF_PLATFORM
    )


def set_rdp_config(content_name):
    rd.get_config().set_param(
        param=f"apis.data.{content_name}.underlying-platform", value=RDP_PLATFORM
    )


@pytest.fixture(
    scope="function",
    params=[set_rdp_config, set_udf_config],
    ids=["rdp-underlying-platform", "udf-underlying-platform"],
)
def set_underlying_platform_config(request):
    content_name = request.instance.pytestmark[0].args[0]
    request.param(content_name)
    underlying_platform_type = request.node.callspec.id[:23]
    yield underlying_platform_type


@pytest.fixture(scope="function", params=["desktop_session"], ids=["desktop_session"])
def setup_app_key():
    ek.set_app_key(desktop_app_key)


@pytest.fixture(name="create_instrument")
def create_unique_instrument():
    instruments = []

    def create_instrument(
        session=None, type_=None, formula=None, udc=None, basket=None
    ):
        universe_name = str(uuid.uuid4())[:8]
        if type_ is None and formula is None:
            formula = "GBP=*3"
        instrument = custom_instruments.manage.create(
            symbol=universe_name,
            instrument_name="FinTechRate",
            type_=type_,
            formula=formula,
            basket=basket,
            udc=udc,
            currency="USD",
            time_zone="UTC",
            description="Story about custom instrument",
            exchange_name="1234",
            session=session,
        )
        symbol = instrument.symbol
        instruments.append((symbol, session))
        return symbol

    yield create_instrument

    for symbol, session in instruments:
        if session is not None:
            session.open()
        response = custom_instruments.search.Definition().get_data(session=session)
        symbols = response.data.df["symbol"].values
        if symbol in symbols:
            custom_instruments.manage.delete(universe=symbol, session=session)


@pytest.fixture
def create_session_with_scope(set_env_base_url):
    tmp = {}

    def _create_session_with_scope(scope, acc):
        session = None
        if acc == "machine":
            session = create_platform_session(credential=edp, scope=scope)
        elif acc == "eikon":
            session = create_platform_session(credential=eikon, scope=scope)
        session.open()
        rd.session.set_default(session)
        tmp.update({"session": session})
        return session

    yield _create_session_with_scope
    session = tmp.get("session")
    session.close()



import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data.content._df_build_type import DFBuildType
from refinitiv.data.content._df_builder_factory import (
    get_dfbuilder,
    content_type_by_build_type,
)
from refinitiv.data.delivery._data._data_provider_factory import (
    data_provider_by_data_type,
)


@pytest.mark.parametrize(
    "content_type,build_type",
    [
        (ContentType.DATA_GRID_RDP, DFBuildType.INDEX),
        (ContentType.DATA_GRID_RDP, DFBuildType.DATE_AS_INDEX),
        (ContentType.DATA_GRID_UDF, DFBuildType.INDEX),
        (ContentType.DATA_GRID_UDF, DFBuildType.DATE_AS_INDEX),
    ],
)
def test_get_df_builder(content_type, build_type):
    # when
    dfbuilder = get_dfbuilder(content_type, build_type)

    # then
    assert callable(dfbuilder) is True


def test_get_df_builder_raise_error_if_not_found_content_type():
    content_type = "__test_content_type__"
    # then
    with pytest.raises(
        ValueError, match="Cannot find mapping for content_type=__test_content_type__"
    ):
        # when
        get_dfbuilder(content_type, DFBuildType.INDEX)


def test_get_df_builder_raise_error_if_not_found_build_type():
    build_type = "__test_build_type__"
    # then
    with pytest.raises(
        ValueError, match="Cannot find mapping for dfbuild_type=__test_build_type__"
    ):
        # when
        get_dfbuilder(ContentType.DATA_GRID_UDF, build_type)



from refinitiv.data._qpl import TenorBidAsk


def test_tenor_bid_ask():
    ask = 1.10
    bid = 2.20
    tenor = "tenor"

    tenor_bid_ask = TenorBidAsk(ask=ask, bid=bid, tenor=tenor)

    assert tenor_bid_ask.ask == ask
    assert tenor_bid_ask.bid == bid
    assert tenor_bid_ask.tenor == tenor

    assert tenor_bid_ask.get_dict() == {"ask": ask, "bid": bid, "tenor": tenor}



import os

import allure
import pytest

import refinitiv.data as rd
from refinitiv.data.content import esg
from refinitiv.data.content.esg import bulk
from tests.integration.content.esg.bulk.helpers import (
    check_folder_contains_init_and_delta_files,
    check_log_file_has_success_download_lines,
    check_log_file_has_success_extract_lines,
    check_log_file_has_no_extract_lines,
    check_downloads_folder,
    check_sqlite_db_data,
    check_folder_cleaned_up,
    check_sqlite_db_is_empty,
)

path = rd.get_config().get("bulk.esg.standard_scores.package.download.path")


@allure.suite("ESG Bulk")
@allure.feature("ESG Bulk")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("esg")
class TestEsgBulk:
    @allure.title(
        "Download all files (init + delta) to folder which does not exist - relative path"
    )
    @pytest.mark.caseid("35564218")
    @pytest.mark.xdist_group(name="esg_group")
    def test_download_all_files_to_folder_which_does_not_exist_relative_path(
        self,
        load_config,
        prepare_file_manager_and_clean_old_files,
        open_platform_session,
    ):
        file_manager = prepare_file_manager_and_clean_old_files
        file_manager.update_files()
        check_downloads_folder(path)
        check_folder_contains_init_and_delta_files(path)
        check_log_file_has_success_download_lines(path)
        check_log_file_has_success_extract_lines(path)

    @allure.title(
        "Update db when local db is empty and init+delta files are already downloaded & unzipped"
    )
    @pytest.mark.caseid("35776180")
    @pytest.mark.dependency(
        depends=["test_download_all_files_to_folder_which_does_not_exist_relative_path"]
    )
    @pytest.mark.xdist_group(name="esg_group")
    def test_update_db_when_local_db_is_empty_and_init_and_delta_files_unzipped(
        self,
        load_config,
        open_platform_session,
    ):
        file_manager = bulk.PackageManager("esg.standard_scores")
        if not os.path.exists(path) or len(os.listdir(path)) == 0:
            file_manager.update_files()
        file_manager.update_db()
        check_sqlite_db_data()

    @allure.title("Get data from db when db is populated")
    @pytest.mark.caseid("35776180")
    @pytest.mark.dependency(
        depends=[
            "test_update_db_when_local_db_is_empty_and_init_and_delta_files_unzipped"
        ]
    )
    @pytest.mark.xdist_group(name="esg_group")
    def test_get_db_data_when_db_is_populated(self, load_config):
        response = esg.standard_scores.Definition(universe="5000268022").get_db_data()
        assert response.data.raw, f"Empty response.data.raw received"
        assert not response.data.df.empty, f"Empty response.data.df received"

    @allure.title("Cleanup db when db contains data")
    @pytest.mark.caseid("36865382")
    @pytest.mark.dependency(
        depends=[
            "test_update_db_when_local_db_is_empty_and_init_and_delta_files_unzipped"
        ]
    )
    @pytest.mark.xdist_group(name="esg_group")
    def test_clean_up_db_when_db_contains_data(self, load_config):
        file_manager = bulk.PackageManager("esg.standard_scores")
        file_manager.cleanup_db()
        check_sqlite_db_is_empty()

    @allure.title("Reset files when downloads folder contains previous files")
    @pytest.mark.caseid("36865099")
    @pytest.mark.dependency(
        depends=["test_download_all_files_to_folder_which_does_not_exist_relative_path"]
    )
    @pytest.mark.xdist_group(name="esg_group")
    def test_reset_files_when_downloads_folder_contains_previous_files(
        self,
        load_config,
        open_platform_session,
    ):
        file_manager = bulk.PackageManager("esg.standard_scores")
        file_manager.reset_files()
        check_downloads_folder(path)
        check_folder_contains_init_and_delta_files(path)
        check_log_file_has_success_download_lines(path)
        check_log_file_has_success_extract_lines(path)

    @allure.title("Download files with auto-extract = false")
    @pytest.mark.caseid("35565422")
    @pytest.mark.xdist_group(name="esg_group")
    def test_download_all_files_with_auto_extract_false(
        self,
        load_config,
        set_auto_extract_to_false,
        prepare_file_manager_and_clean_old_files,
        open_platform_session,
    ):
        file_manager = prepare_file_manager_and_clean_old_files
        file_manager.update_files()
        check_downloads_folder(path)
        check_folder_contains_init_and_delta_files(path)
        check_log_file_has_success_download_lines(path)
        check_log_file_has_no_extract_lines(path)

    @allure.title("Update db when no init file was downloaded -> exception raised")
    @pytest.mark.caseid("35776181")
    @pytest.mark.xdist_group(name="esg_group")
    def test_update_db_when_no_init_file_was_downloaded(
        self, load_config, prepare_file_manager_and_clean_old_files
    ):
        file_manager = prepare_file_manager_and_clean_old_files
        with pytest.raises(FileNotFoundError):
            file_manager.update_db()

    @allure.title("Clean-up files when files were downloaded")
    @pytest.mark.caseid("36397732")
    @pytest.mark.xdist_group(name="esg_group")
    def test_cleanup_files_when_files_were_downloaded(self):
        file_manager = bulk.PackageManager("esg.standard_scores")
        if not os.path.exists(path) or len(os.listdir(path)) == 0:
            file_manager.update_files()
        file_manager.cleanup_files()
        check_folder_cleaned_up(path)



import asyncio
import re
from http import HTTPStatus
from http.server import HTTPServer, BaseHTTPRequestHandler
import pytest
import time
from functools import partial
from threading import Thread

from httpx import Client, AsyncClient

from refinitiv.data._core.session._retry_transport import (
    RetryAsyncTransport,
    RetryTransport,
    RequestRetryException,
    RequestTimeout,
    RequestAttemptsExhausted,
)


class NoLogHTTPHandler(BaseHTTPRequestHandler):
    def log_message(self, format, *args):
        return


class SuccessHTTPHandler(NoLogHTTPHandler):
    attempts_counter = 1

    def __init__(self, answer_at_attempt: int, *args, **kwargs):
        self.answer_at_attempt = answer_at_attempt
        super().__init__(*args, **kwargs)

    def do_GET(self):
        if SuccessHTTPHandler.attempts_counter >= self.answer_at_attempt:
            self.send_response(HTTPStatus.OK)
        else:
            self.send_response(HTTPStatus.NOT_IMPLEMENTED)
        self.end_headers()
        SuccessHTTPHandler.attempts_counter += 1


_err_still_in_flight_pattern = re.compile(
    r"The connection pool was closed while [0-9]+ HTTP requests/responses were still in-flight."
)


@pytest.fixture(scope="module")
def mock_server(handler=NoLogHTTPHandler):
    # Passing 0 to server port argument claims a random free port
    server = HTTPServer(("127.0.0.1", 0), handler)
    server_thread = Thread(
        target=partial(server.serve_forever, poll_interval=0.1), daemon=True
    )
    server_thread.start()
    yield server
    server.shutdown()


def test_asynchronous_execution(mock_server):
    async def create_client(server_port: int):
        with pytest.raises(RequestRetryException):
            async_transport = RetryAsyncTransport(
                total_timeout=10,
                total_attempts=20,
                on_statuses=[501],
                on_methods=["GET"],
                backoff_factor=0.1,
            )
            async with AsyncClient(transport=async_transport) as async_client:
                await async_client.get(f"http://127.0.0.1:{server_port}")

    async def async_run(server_port: int):
        task = asyncio.create_task(create_client(server_port))
        # Check in a loop to make sure there is no small synchronous code in transport.
        # Synchronous code that takes less than 0.25 seconds may be skipped
        for _ in range(3):
            t_start = time.perf_counter()
            await asyncio.sleep(0.2)
            t_finish = time.perf_counter()
            assert t_finish - t_start < 0.25
        task.cancel()
        await asyncio.wait({task})

    mock_server.RequestHandlerClass = NoLogHTTPHandler
    asyncio.run(async_run(mock_server.server_port))


def test_retry_success_async(mock_server):
    async def async_run(server_port: int, succeed_on_attempt: int):
        async_transport = RetryAsyncTransport(
            total_timeout=50,
            total_attempts=10,
            on_statuses=[501],
            on_methods=["GET"],
            backoff_factor=0,
        )
        async_client = AsyncClient(transport=async_transport)
        response = await async_client.get(f"http://127.0.0.1:{server_port}")

        try:
            await async_client.aclose()
        except RuntimeError as e:
            assert _err_still_in_flight_pattern.match(str(e)), str(e)

        assert response.status_code == 200
        assert (
            async_transport._retrying.statistics["attempt_number"] == succeed_on_attempt
        )

    for successful_attempt in [1, 2, 5]:
        handler = partial(
            SuccessHTTPHandler, successful_attempt
        )  # Needed to pass custom argument to the handler
        SuccessHTTPHandler.attempts_counter = 1

        mock_server.RequestHandlerClass = handler
        asyncio.run(async_run(mock_server.server_port, successful_attempt))


def test_timeout_async(mock_server):
    async def async_run(server_port: int):
        timeout = 0.1

        with pytest.raises(RequestTimeout) as error:
            async_transport = RetryAsyncTransport(
                total_timeout=timeout,
                total_attempts=20,
                on_statuses=[501],
                on_methods=["GET"],
                backoff_factor=0.025,
            )
            async_client = AsyncClient(transport=async_transport)
            await async_client.get(f"http://127.0.0.1:{server_port}")

            try:
                await async_client.aclose()
            except RuntimeError as e:
                assert _err_still_in_flight_pattern.match(str(e)), str(e)

        assert timeout <= error.value.total_time

    mock_server.RequestHandlerClass = NoLogHTTPHandler
    asyncio.run(async_run(mock_server.server_port))


def test_attempts_exceeded_async(mock_server):
    async def async_run(server_port: int):
        for attempt_count in [1, 2, 5]:
            with pytest.raises(RequestAttemptsExhausted) as error:
                async_transport = RetryAsyncTransport(
                    total_timeout=50,
                    total_attempts=attempt_count,
                    on_statuses=[501],
                    on_methods=["GET"],
                    backoff_factor=0,
                )
                async_client = AsyncClient(transport=async_transport)
                await async_client.get(f"http://127.0.0.1:{server_port}")

                try:
                    await async_client.aclose()
                except RuntimeError as e:
                    assert _err_still_in_flight_pattern.match(str(e)), str(e)

            assert attempt_count == error.value.attempts_count

    mock_server.RequestHandlerClass = NoLogHTTPHandler
    asyncio.run(async_run(mock_server.server_port))


def test_retry_success(mock_server):
    for successful_attempt in [1, 2, 5]:
        handler = partial(
            SuccessHTTPHandler, successful_attempt
        )  # Needed to pass custom argument to the handler
        SuccessHTTPHandler.attempts_counter = 1
        mock_server.RequestHandlerClass = handler
        transport = RetryTransport(
            total_timeout=50,
            total_attempts=10,
            on_statuses=[501],
            on_methods=["GET"],
            backoff_factor=0,
        )
        client = Client(transport=transport)
        response = client.get(f"http://127.0.0.1:{mock_server.server_port}")

        try:
            client.close()
        except RuntimeError as e:
            assert _err_still_in_flight_pattern.match(str(e)), str(e)

        assert response.status_code == 200
        assert transport._retrying.statistics["attempt_number"] == successful_attempt


def test_timeout(mock_server):
    timeout = 0.1
    mock_server.RequestHandlerClass = NoLogHTTPHandler
    with pytest.raises(RequestTimeout) as error:
        transport = RetryTransport(
            total_timeout=timeout,
            total_attempts=20,
            on_statuses=[501],
            on_methods=["GET"],
            backoff_factor=0.025,
        )
        client = Client(transport=transport)
        client.get(f"http://127.0.0.1:{mock_server.server_port}")

        try:
            client.close()
        except RuntimeError as e:
            assert _err_still_in_flight_pattern.match(str(e)), str(e)

    assert timeout <= error.value.total_time


def test_attempts_exceeded(mock_server):
    mock_server.RequestHandlerClass = NoLogHTTPHandler
    for attempt_count in [1, 2, 5]:
        with pytest.raises(RequestAttemptsExhausted) as error:
            transport = RetryTransport(
                total_timeout=50,
                total_attempts=attempt_count,
                on_statuses=[501],
                on_methods=["GET"],
                backoff_factor=0,
            )
            client = Client(transport=transport)
            client.get(f"http://127.0.0.1:{mock_server.server_port}")

            try:
                client.close()
            except RuntimeError as e:
                assert _err_still_in_flight_pattern.match(str(e)), str(e)

        assert attempt_count == error.value.attempts_count



import refinitiv.data as rd
from refinitiv.data._core.session import set_default
from tests.unit.conftest import StubSession
from tests.unit.fincoders_layer.data_for_tests import (
    SEARCH_QUERY_CFO_RESPONSE,
    SEARCH_QUERY_IBM_RESPONSE,
)


def test_search_query_cfo():
    expected_srt = (
        "  BusinessEntity                                                                              DocumentTitle       PermID           PI\n"
        "0         PERSON           Amy E. Hood - Microsoft Corp - Chief Financial Officer, Executive Vice President  34415553383  34415553383\n"
        "1         PERSON                  Luca Maestri - Apple Inc - Chief Financial Officer, Senior Vice President  34414554748  34414554748\n"
        "2         PERSON        Brian T. Olsavsky - Amazon.com Inc - Chief Financial Officer, Senior Vice President  34417610894  34417610894\n"
        "3         PERSON              Ruth M. Porat - Alphabet Inc - Chief Financial Officer, Senior Vice President  34413960665  34413960665\n"
        "4         PERSON                             David M. Wehner - Meta Platforms Inc - Chief Financial Officer  34414804241  34414804241\n"
        "5         PERSON  Marc D. Hamburg - Berkshire Hathaway Inc - Chief Financial Officer, Senior Vice President  34413152672  34413152672\n"
        "6         PERSON                                       Andrew K. Klatt - Berkshire Hathaway Inc - CFO & COO  34414966250  34414966250\n"
        "7         PERSON                              Xu Hong - Alibaba Group Holding Ltd - Chief Financial Officer  34425652371  34425652371\n"
        "8         PERSON            John Lo - Tencent Holdings Ltd - Chief Financial Officer, Senior Vice President  34414907131  34414907131\n"
        "9         PERSON          Vasant M. Prabhu - Visa Inc - Vice Chairman of the Board, Chief Financial Officer  34413340523  34413340523"
    )
    # given
    session = StubSession(is_open=True, response=SEARCH_QUERY_CFO_RESPONSE)
    set_default(session)

    # when
    testing_df = rd.discovery.search(query="cfo", view=rd.discovery.Views.PEOPLE)

    # teardown
    set_default(None)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_srt


def test_search_query_ibm():
    expected_srt = (
        "  BusinessEntity                                                                                               DocumentTitle       PermID         PI       RIC\n"
        "0   ORGANISATION                                                        International Business Machines Corp, Public Company         <NA>      37036      <NA>\n"
        "1   ORGANISATION                                                                               Banco IBM SA, Private Company         <NA>      76208      <NA>\n"
        "2   QUOTExEQUITY                                     International Business Machines Corp, Ordinary Share, NYSE Consolidated  55839165994    1097326       IBM\n"
        "3   ORGANISATION                           Tiers Corporate Bond Backed Certificates Trust Series Ibm 1997 4, Private Company         <NA>   18062670      <NA>\n"
        "4   QUOTExEQUITY              Eurex International Business Machines Equity Future Chain Contract , Equity Future, USD, Eurex  21481052421   48924732   0#IBMF:\n"
        "5   QUOTExEQUITY              Euronext Amsterdam IBM Dividend Future Chain Contracts, Equity Future, USD, Euronext Amsterdam  21612423771  259118763  0#IBMDF:\n"
        "6   QUOTExEQUITY               Eurex International Business Machines Equity Future Continuation 1, Equity Future, USD, Eurex  21481052892   49450681    IBMFc1\n"
        "7   QUOTExEQUITY               Eurex International Business Machines Equity Future Continuation 2, Equity Future, USD, Eurex  21481053949   50092347    IBMFc2\n"
        "8   QUOTExEQUITY  Euronext Amsterdam IBM Single Stock Dividend Future Continuation 1, Equity Future, USD, Euronext Amsterdam  21613372305  260213021   IBMDFc1\n"
        "9   QUOTExEQUITY               Eurex International Business Machines Equity Future Continuation 3, Equity Future, USD, Eurex  21481053950   50092348    IBMFc3"
    )

    # given
    session = StubSession(is_open=True, response=SEARCH_QUERY_IBM_RESPONSE)
    set_default(session)

    # when
    testing_df = rd.discovery.search(query="IBM")

    # teardown
    set_default(None)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_srt



from refinitiv.data.content.ipa.surfaces import cap


def surface_cap_definition():
    definition = cap.Definition(
        surface_tag="USD_Strike__Tenor_",
        underlying_definition=cap.CapSurfaceDefinition(
            instrument_code="USD",
            discounting_type=cap.DiscountingType.OIS_DISCOUNTING,
        ),
        surface_layout=cap.SurfaceLayout(format=cap.Format.MATRIX),
        surface_parameters=cap.CapCalculationParams(
            valuation_date="2020-03-20",
            x_axis=cap.Axis.STRIKE,
            y_axis=cap.Axis.TENOR,
        ),
    )
    return definition


def invalid_surface_cap_definition():
    definition = cap.Definition(
        surface_tag="USD_Strike__Tenor_",
        underlying_definition=cap.CapSurfaceDefinition(
            instrument_code="USD",
            discounting_type=cap.DiscountingType.OIS_DISCOUNTING,
        ),
        surface_layout=cap.SurfaceLayout(format=cap.Format.MATRIX),
        surface_parameters=cap.CapCalculationParams(
            valuation_date="2020-03-201",
            x_axis=cap.Axis.STRIKE,
            y_axis=cap.Axis.TENOR,
        ),
    )

    return definition



from typing import TYPE_CHECKING

from ._content_response_factory import ContentResponseFactory

if TYPE_CHECKING:
    from ..delivery._data._data_provider import ParsedData

error_message_by_code = {
    "default": "{error_message}. Requested ric: {rics}. Requested fields: {fields}",
    "TS.Intraday.UserRequestError.90001": "{rics} - The universe is not found",
    "TS.Intraday.Warning.95004": "{rics} - Trades interleaving with corrections is currently not supported. Corrections will not be returned.",
    "TS.Intraday.UserRequestError.90006": "{error_message} Requested ric: {rics}",
}


class HistoricalResponseFactory(ContentResponseFactory):
    def get_raw(self, data: "ParsedData") -> dict:
        return data.content_data[0] if data.content_data else {}

    def create_success(self, parsed_data: "ParsedData", **kwargs):
        self._try_write_error(parsed_data, **kwargs)
        return super().create_success(parsed_data, **kwargs)

    def create_fail(self, parsed_data: "ParsedData", **kwargs):
        self._try_write_error(parsed_data, **kwargs)
        return super().create_fail(parsed_data, **kwargs)

    def _try_write_error(
        self, parsed_data: "ParsedData", universe=None, fields=None, **kwargs
    ):
        raw = self.get_raw(parsed_data)
        error_code = parsed_data.first_error_code or raw.get("status", {}).get("code")

        if not error_code:
            return

        error_message = parsed_data.first_error_message or raw.get("status", {}).get(
            "message"
        )
        rics = raw.get("universe", {}).get("ric", universe)
        parsed_data.error_codes = error_code

        if error_code not in error_message_by_code.keys():
            parsed_data.error_messages = error_message_by_code["default"].format(
                error_message=error_message, rics=rics, fields=fields
            )

        else:
            parsed_data.error_messages = error_message_by_code[error_code].format(
                rics=rics, error_message=error_message
            )



import asyncio
import threading
import time
from typing import TYPE_CHECKING, Optional, Callable

from .._content_provider_layer import ContentProviderLayer
from ..._core.session import get_valid_session
from ...delivery._data._data_provider import emit_event
from ...delivery._data._data_provider_factory import (
    get_base_url,
    get_api_config,
    get_url,
)
from ...delivery._data._endpoint_data import RequestMethod, Error
from ..._tools import urljoin

if TYPE_CHECKING:
    from ...delivery._data._data_provider import Response
    from ..._configure import _RDPConfig
    from ..._core.session import Session

DELAY_BEFORE_FIRST_GET_ASYNC_OPERATION = 0.3
DELAY_BETWEEN_TWO_GET_ASYNC_OPERATION = 5


class IPAContentProviderLayer(ContentProviderLayer):
    def __init__(self, content_type, **kwargs):
        super().__init__(content_type, **kwargs)
        self._lock: Optional[threading.Lock] = None
        self._lock_async: Optional[asyncio.Lock] = None
        self._async_url: Optional[str] = None
        self._operation_url: Optional[str] = None
        self._resource_url: Optional[str] = None
        self._delay_before_first_get_operation: Optional[int] = None
        self._delay_between_two_get_operation: Optional[int] = None
        self._config: Optional["_RDPConfig"] = None
        self._session: Optional["Session"] = None

    def _read_config(self):
        base_url = get_base_url(self._content_type, self._config)
        api_config = get_api_config(self._content_type, self._config)
        self._operation_url = urljoin(
            base_url, api_config.get("endpoints.async-operation")
        )
        self._resource_url = urljoin(
            base_url, api_config.get("endpoints.async-resource")
        )
        self._delay_before_first_get_operation = api_config.get(
            "delay-before-first-get-async-operation",
            DELAY_BEFORE_FIRST_GET_ASYNC_OPERATION,
        )
        self._delay_between_two_get_operation = api_config.get(
            "delay-between-two-get-async-operation",
            DELAY_BETWEEN_TWO_GET_ASYNC_OPERATION,
        )

    def _check_initial_response(self, initial_response: "Response") -> bool:
        check_failed = False

        # At this step, normal response should have 202 status code,
        # "Accepted" error message and location header
        status_code = initial_response.http_status.get("http_status_code")
        if status_code != 202:
            check_failed = True
            first_error = initial_response.errors[0]
            err_msg = first_error.message
            err_msg = (
                f"Async IPA response "
                f"status code {status_code}|{err_msg} != 202|Accepted"
            )
            self._session.error(err_msg)
            self._write_error(initial_response, first_error.code, err_msg)

        location = initial_response.http_headers.get("location")
        if not location:
            check_failed = True
            err_msg = (
                "IPA Asynchronous request operation failed, "
                "response doesn't contain location."
            )
            self._session.error(err_msg)
            self._write_error(initial_response, 0, err_msg)

        return check_failed

    def _check_operation_response(self, operation_response: "Response") -> bool:
        check_failed = False
        resource_location = operation_response.data.raw.get("resourceLocation")
        if not resource_location:
            check_failed = True
            err_msg = (
                "IPA Asynchronous request resource failed, "
                "operation response doesn't contain resource location."
            )
            self._session.error(err_msg)
            self._write_error(operation_response, 0, err_msg)

        return check_failed

    def _write_error(self, response: "Response", err_code: int, err_msg: str):
        error = Error(err_code, err_msg)
        response.errors.append(error)
        response.is_success = False

    def _raise_error_no_async_url(self):
        raise AttributeError(
            f"Asynchronous endpoint is not available for this content provider"
        )

    def get_data(
        self,
        session: Optional["Session"] = None,
        on_response: Optional[Callable] = None,
        async_mode: Optional[bool] = None,
    ):
        self._session = get_valid_session(session)
        self._config: "_RDPConfig" = self._session.config
        self._async_url = get_url(
            self._content_type, self._config, request_mode="async"
        )
        response = None

        if async_mode and self._async_url:
            self._lock = threading.Lock()
            self._read_config()
            response = self._get_data_with_async_mode()
            on_response and emit_event(on_response, response, self, session)
            self._check_response(response, self._config)

        elif async_mode and not self._async_url:
            self._raise_error_no_async_url()

        else:
            response = super().get_data(session, on_response)

        return response

    def _get_data_with_async_mode(self) -> "Response":
        initial_response = self._provider.get_data(
            self._session, self._async_url, **self._kwargs
        )

        check_failed = self._check_initial_response(initial_response)
        if check_failed:
            return initial_response

        operation_response = None
        location = initial_response.http_headers.get("location")
        operation_id = location.rsplit("/", 1)[-1]
        url = urljoin(self._operation_url, operation_id)
        loop_retrieving = True
        while loop_retrieving:
            # wait before requesting operation status
            time.sleep(self._delay_before_first_get_operation)
            # while request operation not succeeded or not failed, request operation

            with self._lock:
                self._session.debug(f"Request operation :\n {url}")
                operation_response = self._provider.get_data(
                    self._session, url, method=RequestMethod.GET, **self._kwargs
                )

            status_code = operation_response.http_status.get("http_status_code")
            if status_code != 200:
                # operation status should be 200, otherwise,
                # it failed, then return response as an error
                return operation_response

            status_text = operation_response.data.raw.get("status")
            if status_text in {"failed", "succeeded"}:
                # request succeeded or failed, in both cases,
                # stop to wait and retrieve result
                loop_retrieving = False
            else:
                # wait for 5 sec before next request
                time.sleep(self._delay_between_two_get_operation)

        check_failed = self._check_operation_response(operation_response)
        if check_failed:
            return operation_response

        with self._lock:
            resource_location = operation_response.data.raw.get("resourceLocation")
            resource_id = resource_location.rsplit("/", 1)[-1]
            url = urljoin(self._resource_url, resource_id)
            self._session.debug(f"Request resource :\n {url}")
            resource_response = self._provider.get_data(
                self._session, url, method=RequestMethod.GET, **self._kwargs
            )

        return resource_response

    async def get_data_async(
        self,
        session: Optional["Session"] = None,
        on_response: Optional[Callable] = None,
        async_mode: Optional[bool] = None,
    ):
        self._session = get_valid_session(session)
        self._config: "_RDPConfig" = self._session.config
        self._async_url = get_url(
            self._content_type, self._config, request_mode="async"
        )
        response = None

        if async_mode and self._async_url:
            self._lock_async = asyncio.Lock()
            self._read_config()
            response = await self._get_data_with_async_mode_async()
            on_response and emit_event(on_response, response, self, session)
            self._check_response(response, self._config)

        elif async_mode and not self._async_url:
            self._raise_error_no_async_url()

        else:
            response = await super().get_data_async(session, on_response)

        return response

    async def _get_data_with_async_mode_async(self):
        initial_response = await self._provider.get_data_async(
            self._session, self._async_url, **self._kwargs
        )

        check_failed = self._check_initial_response(initial_response)
        if check_failed:
            return initial_response

        operation_response = None
        location = initial_response.http_headers.get("location")
        operation_id = location.rsplit("/", 1)[-1]
        url = urljoin(self._operation_url, operation_id)
        loop_retrieving = True
        while loop_retrieving:
            # wait before requesting operation status
            await asyncio.sleep(self._delay_before_first_get_operation)
            # while request operation not succeeded or not failed, request operation

            async with self._lock_async:
                self._session.debug(f"Request operation :\n {url}")
                operation_response = await self._provider.get_data_async(
                    self._session, url, method=RequestMethod.GET, **self._kwargs
                )

            status_code = operation_response.http_status.get("http_status_code")
            if status_code != 200:
                # operation status should be 200, otherwise,
                # it failed, then return response as an error
                return operation_response

            status_text = operation_response.data.raw.get("status")
            if status_text in {"failed", "succeeded"}:
                # request succeeded or failed, in both cases,
                # stop to wait and retrieve result
                loop_retrieving = False
            else:
                # wait for 5 sec before next request
                await asyncio.sleep(self._delay_between_two_get_operation)

        check_failed = self._check_operation_response(operation_response)
        if check_failed:
            return operation_response

        async with self._lock_async:
            resource_location = operation_response.data.raw.get("resourceLocation")
            resource_id = resource_location.rsplit("/", 1)[-1]
            url = urljoin(self._resource_url, resource_id)
            self._session.debug(f"Request resource :\n {url}")
            resource_response = await self._provider.get_data_async(
                self._session, url, method=RequestMethod.GET, **self._kwargs
            )

        return resource_response



# coding: utf8

__all__ = ["DayOfWeek"]

from enum import Enum, unique


@unique
class DayOfWeek(Enum):
    """
    The day of week to which dates are adjusted.
    The first date in the list is defined as corresponding day of week following the start date.
    The last date in the list is defined as corresponding day of week preceding the end date.

    The possible values are:
        - Sunday,
        - Monday,
        - Tuesday,
        - Wednesday,
        - Thursday,
        - Friday,
        - Saturday
    """

    SUNDAY = "Sunday"
    MONDAY = "Monday"
    TUESDAY = "Tuesday"
    WEDNESDAY = "Wednesday"
    THURSDAY = "Thursday"
    FRIDAY = "Friday"
    SATURDAY = "Saturday"



import os
import shutil
import pytest

from refinitiv.data import _configure as configure
from refinitiv.data.content.esg.bulk._package_manager import (
    get_console_logger,
    create_logpath,
    _PackageManager,
    get_package_config_by_package_name,
)
from tests.unit.conftest import StubSession


def test_get_console_logger():
    # given
    test_id = "1"

    # when
    result = get_console_logger(test_id)

    # then
    assert result


def test_create_log_path():
    # given
    package_name = "esg.test_package"
    configure.set_param(
        f"bulk.{package_name}.package.download.path",
        "test_folder",
        auto_create=True,
    )

    # when
    result = create_logpath(package_name)

    # then
    assert result == os.path.join("test_folder", "log.txt")

    os.rmdir("test_folder")


def test_file_manager():
    try:
        # given
        package_manager = _PackageManager("esg.standard_scores")
        # when
        fm = package_manager.file_manager
        # then
        assert fm is package_manager.file_manager
        shutil.rmtree("downloads")
    except OSError:
        pass


def test_get_package_config_by_package_name():
    # given
    package_name = "wrong_name"

    # then
    with pytest.raises(ValueError):
        # when
        get_package_config_by_package_name(package_name)


def test_file_manager_get_session():
    session = StubSession()
    package_manager = _PackageManager("esg.standard_scores", session)
    assert isinstance(package_manager.file_manager.session, StubSession)



from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    CombinedShift,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves._enums import (
    ShiftType,
    ShiftUnit,
)


def test_combined_shift():
    # given
    expected_dict = {
        "amount": 10.0,
        "shiftType": "Additive",
        "shiftUnit": "Absolute",
        "endTenor": "string",
        "startTenor": "string",
    }

    # when
    testing_obj = CombinedShift(
        amount=10.0,
        shift_type=ShiftType.ADDITIVE,
        shift_unit=ShiftUnit.ABSOLUTE,
        end_tenor="string",
        start_tenor="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.amount == 10.0
    assert testing_obj.shift_type == ShiftType.ADDITIVE
    assert testing_obj.shift_unit == ShiftUnit.ABSOLUTE
    assert testing_obj.end_tenor == "string"
    assert testing_obj.start_tenor == "string"



from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    LongEndShift,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves._enums import (
    ShiftType,
    ShiftUnit,
)


def test_long_end_shift():
    # given
    expected_dict = {"amount": 10.0, "shiftType": "Additive", "shiftUnit": "Absolute"}

    # when
    testing_obj = LongEndShift(
        amount=10.0,
        shift_type=ShiftType.ADDITIVE,
        shift_unit=ShiftUnit.ABSOLUTE,
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.amount == 10.0
    assert testing_obj.shift_type == ShiftType.ADDITIVE
    assert testing_obj.shift_unit == ShiftUnit.ABSOLUTE



import pytest
from mock.mock import MagicMock


@pytest.fixture(scope="function")
def session_mock():
    session_mock_ = MagicMock()
    config_mock = MagicMock()
    config_mock.get_str.return_value = "string"
    session_mock_._get_endpoint_config.return_value = config_mock
    return session_mock_



__all__ = (
    "Definition",
    "Definitions",
    "DateMovingConvention",
    "EndOfMonthConvention",
    "HolidayOutputs",
)

from ._definition import Definition, Definitions
from ..._enums import DateMovingConvention, EndOfMonthConvention, HolidayOutputs



import allure
import pytest

from refinitiv.data.content.ipa.surfaces import fx
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.conftest import (
    check_surface,
    check_surface_curve,
    check_surface_point,
)
from tests.integration.content.ipa.contracts.conftest import (
    check_http_status_is_success_and_df_value_not_empty,
)
from tests.integration.content.ipa.surfaces.fx.conftest import (
    surface_fx_definition,
    invalid_surface_fx_definition,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
)


@allure.suite("Content object - Surface Fx")
@allure.feature("Content object - Surface Fx")
@allure.severity(allure.severity_level.CRITICAL)
class TestSurfaceFx:
    @allure.title(
        "Create a surface fx definition object with valid params and get data"
    )
    @pytest.mark.caseid("37821399")
    def test_surface_fx_with_valid_params_and_get_data(self, open_session):
        response = surface_fx_definition().get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        surface = response.data.surface
        check_surface(surface)
        curve = surface.get_curve("2021-08-19", fx.Axis.X)
        print(curve)
        check_surface_curve(curve)
        point = surface.get_point("2021-08-19", 1.109)
        check_surface_point(point)

    @allure.title(
        "Create a surface fx definition object with valid params and get data asynchronously"
    )
    @pytest.mark.caseid("37821400")
    @pytest.mark.smoke
    async def test_surface_fx_with_valid_params_and_get_data_async(self, open_session_async):
        valid_response, invalid_response = await get_async_response_from_definitions(
            surface_fx_definition(), invalid_surface_fx_definition()
        )
        check_response_status(valid_response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(valid_response)
        surface = valid_response.data.surface
        check_surface(surface)
        curve = surface.get_curve("2020-08-20", fx.Axis.X)
        check_surface_curve(curve)
        point = surface.get_point("2021-11-19", 1.12)
        check_surface_point(point)

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="VolSurf.10300",
            expected_error_message="The service failed to build the volatility surface. (Additional technical information: Invalid input: CrossCurrency value is unknown. \'INVAL\' is not a valid FxCrossCode.)",
        )

    @pytest.mark.xfail(reason="https://jira.refinitiv.com/browse/EAPI-4119")
    @allure.title("Create a surface fx - short definition")
    @pytest.mark.caseid("37821401")
    def test_surface_fx_with_simple_definition(self, open_session):
        response = fx.Definition(underlying_definition="EURUSD").get_data()
        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title("Create a invalid surface fx definition and get RDError")
    @pytest.mark.caseid("39606894")
    def test_invalid_surface_fx_definition(self, open_session):
        with pytest.raises(
            RDError,
            match="Error code VolSurf.10300 | The service failed to build the volatility surface",
        ):
            invalid_surface_fx_definition().get_data()



# coding: utf8

from enum import Enum, unique


@unique
class StubRule(Enum):
    ISSUE = "Issue"
    LONG_FIRST_FULL = "LongFirstFull"
    MATURITY = "Maturity"
    SHORT_FIRST_FULL = "ShortFirstFull"
    SHORT_FIRST_PRO_RATA = "ShortFirstProRata"
    SHORT_LAST_PRO_RATA = "ShortLastProRata"



from pathlib import Path

import pytest

from refinitiv.data.delivery.cfs import file_downloader
from refinitiv.data.delivery.cfs._file_downloader_facade import FileDownloader
from tests.unit.conftest import StubSession


def test_call_download_without_path():
    # given
    input_value = "filename.json.gzip"
    expected_value = Path("filename.json.gzip")
    file_downloader = FileDownloader(url="", filename_ext=input_value)

    # when
    file_downloader.download()
    testing_value = Path(file_downloader._downloaded_filepath)

    # then
    assert testing_value == expected_value


def test_call_download_with_pass_path():
    # given
    input_value = "filename.json.gzip"
    expected_value = Path("path/to/file/filename.json.gzip")
    file_downloader = FileDownloader(url="", filename_ext=input_value)

    # when
    file_downloader.download("path/to/file")
    testing_value = Path(file_downloader._downloaded_filepath)

    # then
    assert testing_value == expected_value


def test_call_extract_without_path():
    # given
    input_value = "filename.json.gzip"
    expected_value = "filename.json"
    file_downloader = FileDownloader(url="", filename_ext=input_value)

    # when
    testing_value = file_downloader.extract()

    # then
    assert testing_value == expected_value


def test_call_extract_with_pass_path():
    # given
    input_value = "filename.json.gzip"
    expected_value = Path("path/to/file/filename.json")
    file_downloader = FileDownloader(url="", filename_ext=input_value)

    # when
    testing_value = file_downloader.extract("path/to/file")
    testing_value = Path(testing_value)

    # then
    assert testing_value == expected_value


def test_call_download_then_call_extract_methods():
    # given
    input_value = "filename.json.gzip"
    expected_value = "filename.json"
    file_downloader = FileDownloader(url="", filename_ext=input_value)

    # when
    testing_value = file_downloader.download().extract()

    # then
    assert testing_value == expected_value


def test_call_download_with_path_then_call_extract_with_path():
    # given
    input_value = "filename.json.gzip"
    expected_value = Path("path/to/extract/filename.json")
    file_downloader = FileDownloader(url="", filename_ext=input_value)

    # when
    testing_value = file_downloader.download("path/to/download").extract(
        "path/to/extract"
    )
    testing_value = Path(testing_value)

    # then
    assert testing_value == expected_value


def test_call_download_with_path_then_call_extract():
    # given
    input_value = "filename.json.gzip"
    expected_value = Path("path/to/download/filename.json")
    file_downloader = FileDownloader(url="", filename_ext=input_value)

    # when
    testing_value = file_downloader.download("path/to/download").extract()
    testing_value = Path(testing_value)

    # then
    assert testing_value == expected_value


def test_call_download_then_call_extract_with_path():
    # given
    input_value = "filename.json.gzip"
    expected_value = Path("path/to/extract/filename.json")
    file_downloader = FileDownloader(url="", filename_ext=input_value)

    # when
    testing_value = file_downloader.download().extract("path/to/extract")
    testing_value = Path(testing_value)

    # then
    assert testing_value == expected_value


def test_cfs_stream():
    session = StubSession(is_open=True)
    definition = file_downloader.Definition(
        {"id": "file_id", "filename": "filename_ext"}
    )

    with pytest.raises(FileNotFoundError, match="file not found"):
        definition.retrieve(session)



from unittest.mock import patch

import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data.delivery._stream._stream_factory import StreamDetails, load_config
from tests.unit.conftest import StubSession


def fake_func(*args):
    return args


@pytest.mark.parametrize(
    ("api_config_str", "expected_config_str"),
    [
        ["streaming.pricing.endpoints.main", "apis.streaming.pricing.endpoints.main"],
        [
            "streaming.pricing.endpoints.main.path",
            "apis.streaming.pricing.endpoints.main",
        ],
        [
            "apis.streaming.pricing.endpoints.main",
            "apis.streaming.pricing.endpoints.main",
        ],
        [
            "apis.streaming.pricing.endpoints.main.path",
            "apis.streaming.pricing.endpoints.main",
        ],
    ],
)
def test_load_config(api_config_str, expected_config_str):
    # given
    session = StubSession()
    stream_details = StreamDetails(
        ContentType.STREAMING_CUSTOM, ..., ..., api_config_str
    )

    # when
    with patch(
        "refinitiv.data.delivery._stream._stream_factory.get_cxn_config",
        wraps=fake_func,
    ) as mock_func:
        load_config(stream_details, session)
        mock_func.assert_called_once_with(expected_config_str, session)


def test_load_config_raise_exception_with_empty_api_config_str():
    # given
    api_config_str = ""
    session = StubSession()
    stream_details = StreamDetails(
        ContentType.STREAMING_CUSTOM, ..., ..., api_config_str
    )

    # when
    with pytest.raises(
        ValueError,
        match="For ContentType.STREAMING_CUSTOM, api_config_key cannot be None",
    ):
        load_config(stream_details, session)


@pytest.mark.parametrize(
    "api_config_str",
    [
        "streaming.pricing.main",
        "apis.streaming.pricing.main",
        "apis.streaming.pricing.endpoints.",
        "apis.streaming.pricing.main.path",
        "streaming.pricing.main.path",
    ],
)
def test_load_config_raise_exception(api_config_str):
    # given
    session = StubSession()
    stream_details = StreamDetails(
        ContentType.STREAMING_CUSTOM, ..., ..., api_config_str
    )

    # when
    with pytest.raises(ValueError, match="Not an existing path"):
        load_config(stream_details, session)


@pytest.mark.parametrize(
    "api_config_str",
    [
        "streaming.pricing.endpoints",
        "apis.streaming.pricing.endpoints",
    ],
)
def test_load_config_raise_exception_not_valid_format(api_config_str):
    # given
    session = StubSession()
    stream_details = StreamDetails(
        ContentType.STREAMING_CUSTOM, ..., ..., api_config_str
    )

    # when
    with pytest.raises(
        ValueError, match="Not a valid format, use `apis.streaming.xxx.endpoints.xxx"
    ):
        load_config(stream_details, session)



# coding: utf8


from enum import Enum, unique


@unique
class InterpolationMode(Enum):
    CUBIC_DISCOUNT = "CubicDiscount"
    CUBIC_RATE = "CubicRate"
    CUBIC_SPLINE = "CubicSpline"
    FORWARD_MONOTONE_CONVEX = "ForwardMonotoneConvex"
    LINEAR = "Linear"
    LOG = "Log"



from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from .._tools._dataframe import convert_df_columns_to_datetime, convert_dtypes

if TYPE_CHECKING:
    from .context_collection import ADCContext, HPContext
    from ._containers import HPAndCustInstDataContainer, ADCDataContainer


def convert_types(column: list, column_names: list) -> list:
    date_columns = [
        i
        for i, column_name in enumerate(column_names)
        if any([i for i in ["Date", "date", "_DT", "DATE"] if i in column_name])
        and all([i if i not in column_name else False for i in ["DateType", "Dates"]])
    ]
    result = [i if i is not None and i != "" else pd.NA for i in column]

    if date_columns:
        for i in date_columns:
            result[i] = pd.to_datetime(result[i])

    return result


class DataProvider:
    def get_df(
        self,
        adc: "ADCContext",
        mixed: "HPContext",
        adc_data: "ADCDataContainer",
        mixed_data: "HPAndCustInstDataContainer",
        use_field_names_in_headers: bool,
    ) -> pd.DataFrame():

        if not mixed_data and not adc_data:
            return pd.DataFrame()

        elif mixed.can_build_df:
            return mixed.df

        elif adc.can_build_df:
            return adc.df

        # merge_data
        columns, data = self.merge_data(
            adc_data, mixed_data, use_field_names_in_headers
        )

        if not any(columns):
            return pd.DataFrame()

        else:
            df = pd.DataFrame(np.array(data), columns=columns)
            df = convert_df_columns_to_datetime(
                df, entry="Date", utc=True, delete_tz=True
            )
            return convert_dtypes(df)

    def merge_data(
        self,
        adc_data: "ADCDataContainer",
        mixed_data: "HPAndCustInstDataContainer",
        use_field_names_in_headers: bool,
    ):
        adc_columns, data = self.get_columns_and_data_from_adc_raw(
            adc_data.raw, use_field_names_in_headers
        )
        columns = adc_columns + mixed_data.columns
        if not adc_columns and mixed_data.columns:
            columns.insert(0, "Instrument")

        result_data = []
        for universe in mixed_data.raw:
            if universe in data:
                for column in data[universe]:
                    column.extend(mixed_data.raw[universe])
                    result_data.append(column)

            else:
                data[universe] = (
                    [universe]
                    + [pd.NA] * (len(adc_columns) - 1)
                    + mixed_data.raw[universe]
                )
                result_data.append(data[universe])

        return columns, result_data

    def get_columns_and_data_from_adc_raw(
        self, raw: dict, use_field_names_in_headers: bool
    ) -> tuple:
        headers = raw.get("headers", [[{}]])

        if headers and isinstance(headers[0], list):
            columns = [
                header.get("displayName")
                if not use_field_names_in_headers
                else header.get("field", header.get("displayName"))
                for header in headers[0]
            ]
        else:
            columns = [
                header.get("name")
                if use_field_names_in_headers
                else header.get("title")
                for header in headers
            ]

            if "instrument" in columns:
                columns[columns.index("instrument")] = "Instrument"

        _data = raw.get("data", [])
        data = {}

        for column in _data:
            if column[0] not in data:
                data[column[0]] = []

            data[column[0]].append(convert_types(column, columns))

        if not data:
            columns = []

        return columns, data



# coding: utf8

from enum import Enum, unique


@unique
class FxLegType(Enum):
    FX_FORWARD = "FxForward"
    FX_NON_DELIVERABLE_FORWARD = "FxNonDeliverableForward"
    FX_SPOT = "FxSpot"
    SWAP_FAR = "SwapFar"
    SWAP_NEAR = "SwapNear"



from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    BidAskFieldsDescription,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    FieldFormulaDescription,
)


def test_bid_ask_fields_description():
    # given
    expected_dict = {"bid": {}, "ask": {}}

    # when
    testing_obj = BidAskFieldsDescription(
        bid=FieldFormulaDescription(),
        ask=FieldFormulaDescription(),
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.bid == FieldFormulaDescription()
    assert testing_obj.ask == FieldFormulaDescription()



import pytest

from refinitiv.data import errors


@pytest.fixture(
    params=[
        attr_name
        for attr_name in dir(errors)
        if not attr_name.startswith("_") and attr_name not in {"ItemWasNotRequested", "ScopeError"}
    ],
)
def error_class(request):
    attr_name = request.param
    cls = getattr(errors, attr_name)
    return cls


def test_error_code(error_class):
    # given
    code = 1

    # when
    error = error_class(code, "")

    # then
    assert error.code == code


def test_error_message(error_class):
    # given
    message = "message"

    # when
    error = error_class(0, message)

    # then
    assert error.message == message


def test_error_to_str(error_class):
    # given
    code = 1
    message = "message"
    expected = f"Error code {code} | {message}"
    error = error_class(code, message)

    # when
    testing_string = str(error)

    # then
    assert testing_string == expected, testing_string



from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    OverrideFxForwardTurn,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    FxForwardTurnFields,
)


def test_override_fx_forward_turn():
    # given
    expected_dict = {
        "startDate": "string",
        "endDate": "string",
        "fields": {},
        "date": "string",
        "turnTag": "string",
    }

    # when
    testing_obj = OverrideFxForwardTurn(
        start_date="string",
        end_date="string",
        fields=FxForwardTurnFields(),
        date="string",
        turn_tag="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.start_date == "string"
    assert testing_obj.end_date == "string"
    assert testing_obj.fields == FxForwardTurnFields()
    assert testing_obj.date == "string"
    assert testing_obj.turn_tag == "string"



from enum import Enum, unique


@unique
class BasisSplineSmoothModel(Enum):
    ANDERSON_SMOOTHING_SPLINE_MODEL = "AndersonSmoothingSplineModel"
    MC_CULLOCH_LINEAR_REGRESSION = "McCullochLinearRegression"
    WAGGONER_SMOOTHING_SPLINE_MODEL = "WaggonerSmoothingSplineModel"



from types import SimpleNamespace

import pytest

import refinitiv.data as rd


@pytest.fixture(scope="function")
def session():
    session = rd.session.desktop.Definition(app_key="test_app_key").get_session()
    yield session
    rd.session.set_default(None)


@pytest.fixture(scope="function")
def stub_session():
    self = SimpleNamespace()
    self.debug = lambda *_: None
    self.error = lambda *_: None
    return self



from datetime import timedelta, datetime

import allure
import pytest

from refinitiv.data.content import custom_instruments as ci
from refinitiv.data.content.custom_instruments._custom_instruments_data_provider import (
    get_user_id,
)
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason, Intervals
from tests.integration.content.custom_instruments.summaries.conftest import (
    get_summaries_definition_with_one_universe,
    get_summaries_definition_with_list_universes,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_non_empty_response_data,
    check_response_status,
    check_response_is_success,
    check_response_values,
    check_universe_in_response,
)


@allure.suite("Content object - Custom Instrument")
@allure.feature("Custom Instrument Summaries")
@allure.severity(allure.severity_level.CRITICAL)
class TestCustomInstrumentSummaries:
    @allure.title(
        "Create Custom Instrument Summaries Definition object using a string/list of strings to define universe and "
        "IntradayInterval"
    )
    @pytest.mark.caseid(["C42991501", "C42991503"])
    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "summaries_definition",
        [
            get_summaries_definition_with_one_universe,
            get_summaries_definition_with_list_universes,
        ],
    )
    def test_create_custom_instrument_summaries_object_with_universe_intraday_interval(
        self, open_desktop_session, summaries_definition, create_instrument
    ):
        definition, universes = summaries_definition(
            create_instrument, ci.Intervals.ONE_MINUTE
        )
        response = definition.get_data()
        check_universe_in_response(response, universes)

    @allure.title(
        "Create Custom Instrument Summaries Definition object using a string/list of strings to define universe and "
        "InterdayInterval"
    )
    @pytest.mark.caseid(["C42993448"])
    @pytest.mark.parametrize(
        "summaries_definition",
        [
            get_summaries_definition_with_one_universe,
            get_summaries_definition_with_list_universes,
        ],
    )
    def test_create_custom_instrument_summaries_object_with_universe_interday_interval(
        self, open_desktop_session, summaries_definition, create_instrument
    ):
        definition, universes = summaries_definition(
            create_instrument, ci.Intervals.DAILY
        )
        response = definition.get_data()
        check_universe_in_response(response, universes)

    @allure.title(
        "Create Custom Instrument Summaries Definition object using some valid optional parameters"
    )
    @pytest.mark.caseid("C42994114")
    @pytest.mark.parametrize(
        "start_date, end_date",
        [
            (timedelta(days=-90, hours=1), timedelta(0)),
            (timedelta(days=-90, hours=1), datetime.now()),
            ("2022-02-06T00:10:00", "2022-05-06T07:10:00"),
        ],
        ids=["timedelta", "datetime", "string"],
    )
    def test_create_custom_instrument_summaries_object_with_valid_optional_parameters(
        self, open_desktop_session, create_instrument, start_date, end_date
    ):
        instrument = create_instrument()
        response = ci.summaries.Definition(
            universe=[instrument],
            interval=ci.Intervals.WEEKLY,
            count=30,
            start=start_date,
            end=end_date,
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_values(
            response,
            expected_universe_ric=instrument,
            expected_interval=Intervals.WEEKLY,
            expected_summary_timestamp_label="endPeriod",
        )

    @allure.title(
        "Create Custom Instrument Summaries Definition object using invalid universe"
    )
    @pytest.mark.caseid("C42994133")
    @pytest.mark.parametrize("universe", ["invalid_universe"])
    def test_create_custom_instrument_summaries_object_with_invalid_universe(
        self, open_desktop_session, universe
    ):
        with pytest.raises(RDError) as error:
            ci.summaries.Definition(universe).get_data()
        assert (
            "Error code 1 | No data to return, please check errors: ERROR: No successful response.\n(404, CustomInstrument with symbol: S)invalid_universe)"
            == str(error.value)
        )

    @allure.title(
        "Create Custom Instrument Summaries Definition object using invalid Interval"
    )
    @pytest.mark.caseid("C42994652")
    @pytest.mark.parametrize("interval", ["invalid_interval"])
    def test_create_custom_instrument_summaries_object_using_invalid_interval(
        self, open_desktop_session, create_instrument, interval
    ):
        instrument = create_instrument()
        with pytest.raises(AttributeError) as error:
            ci.summaries.Definition(universe=instrument, interval=interval).get_data()

        assert (
            str(error.value)
            == f"Value '{interval}' must be in ['PT1M', 'PT5M', 'PT10M', 'PT30M', 'PT60M', 'PT1H', 'P1D', "
            f"'P7D', 'P1W', 'P1M', 'P3M', 'P12M', 'P1Y']"
        )

    @allure.title(
        "Create Custom Instrument Summaries Definition object when session is not open"
    )
    @pytest.mark.caseid("C42995388")
    def test_create_custom_instrument_summaries_object_with_session_not_opened(
        self, open_desktop_session, create_instrument
    ):
        session = open_desktop_session
        session.close()
        with pytest.raises(ValueError) as error:
            ci.summaries.Definition(universe="S)symbol").get_data()
        assert str(error.value) == "Session is not opened. Can't send any request"

    @allure.title(
        "Create Custom Instrument Summaries get data async with valid and invalid universe"
    )
    @pytest.mark.caseid("C42995791")
    @pytest.mark.parametrize("universe", ["invalid_universe"])
    async def test_custom_instrument_summaries_get_data_async(
        self, open_desktop_session_async, create_instrument, universe
    ):
        symbol = create_instrument()
        uuid = get_user_id()
        valid_response, invalid_response = await get_async_response_from_definitions(
            ci.summaries.Definition(symbol),
            ci.summaries.Definition(universe),
        )
        check_response_is_success(valid_response)
        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.FOUR_HUNDRED_FOUR,
            expected_http_reason=HttpReason.NOT_FOUND,
            expected_error_code="404",
            expected_error_message=f"CustomInstrument with symbol: S){universe}.{uuid} for user: {uuid} not found!. "
            f"Requested ric: {universe}. Requested fields: None",
        )



import allure
import pytest

from refinitiv.data.content.ipa import surfaces
from refinitiv.data.content.ipa.surfaces import cap
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.conftest import (
    check_surface,
    check_surface_curve,
    check_surface_point,
)
from tests.integration.content.ipa.surfaces.cap.conftest import (
    surface_cap_definition,
    invalid_surface_cap_definition,
)
from tests.integration.content.ipa.surfaces.eti.conftest import (
    surface_eti_definition,
)
from tests.integration.content.ipa.surfaces.fx.conftest import (
    surface_fx_definition,
    invalid_surface_fx_definition,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    check_non_empty_response_data,
)


@allure.suite("Content object - Surfaces")
@allure.feature("Content object - Surfaces")
@allure.severity(allure.severity_level.CRITICAL)
class TestSurfaces:
    @pytest.mark.xfail(reason="https://jira.refinitiv.com/browse/EAPI-3135")
    @allure.title(
        "Create a surfaces definition object with valid definitions and get data"
    )
    @pytest.mark.caseid("39606902")
    @pytest.mark.smoke
    def test_surfaces_valid_params_and_get_data(self, open_session):
        response = surfaces.Definitions(
            [
                surface_cap_definition(),
                surface_fx_definition(),
                surface_eti_definition(),
            ]
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        surface = response.data.surfaces
        check_surface(surface)
        curve = surface.get_curve("0.5", cap.Axis.X)
        check_surface_curve(curve)
        point = surface.get_point("0.5", "2Y")
        check_surface_point(point)

    @allure.title(
        "Create a surfaces definition object with valid and invalid definitions and get data"
    )
    @pytest.mark.caseid("39606903")
    @pytest.mark.smoke
    def test_surfaces_valid_and_invalid_params_and_get_data(self, open_session):
        response = surfaces.Definitions(
            [
                surface_cap_definition(),
                surface_fx_definition(),
                invalid_surface_fx_definition(),
            ]
        ).get_data()

        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="VolSurf.10300",
            expected_error_message="The service failed to build the volatility surface. (Additional technical information: Invalid input: CrossCurrency value is unknown. 'INVAL' is not a valid FxCrossCode.)",
        )

    @allure.title(
        "Create a surfaces definition object with valid and invalid definitions and get data asynchronously"
    )
    @pytest.mark.caseid("39606904")
    @pytest.mark.smoke
    async def test_surfaces_with_valid_and_invalid_params_and_get_data_async(
        self, open_session_async
    ):
        valid_response, invalid_response = await get_async_response_from_definitions(
            surfaces.Definitions([surface_eti_definition(), surface_fx_definition()]),
            surfaces.Definitions([invalid_surface_cap_definition()]),
        )

        check_response_status(valid_response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="VolSurf.10200",
            expected_error_message="Market data: The service failed to retrieve the necessary data to build the volatility surface",
        )

    @allure.title("Create a invalid surfaces definition and get RDError")
    @pytest.mark.caseid("39606905")
    def test_invalid_surface_cap_definition(self, open_session):
        with pytest.raises(
            RDError,
            match="Error code VolSurf.10200 | Market data: The service failed to retrieve the necessary data to build the volatility surface",
        ):
            surfaces.Definitions([invalid_surface_cap_definition()]).get_data()



import allure
import pytest

from refinitiv.data.content import news
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.news.conftest import (
    check_story_title,
)
from tests.integration.helpers import (
    check_response_status,
    get_async_response_from_definition,
)

ERROR_MESSAGE_PATTERN = {
    "PLATFORM": "Error code 404 | Error while calling the NEP backend: Story not found",
    "DESKTOP": "Error code 404 | Error while calling the NEP backend: Story not found",
}


@allure.suite("Content object - News Story")
@allure.feature("Content object - News Story")
@allure.severity(allure.severity_level.CRITICAL)
@pytest.mark.underlying_content("news")
class TestNewsStory:
    @allure.title("Create News Story object with opened session and valid story ID")
    @pytest.mark.caseid("34114894")
    @pytest.mark.parametrize(
        "id,story_title",
        [
            (
                "urn:newsml:reuters.com:20171030:nTOPMTL:169",
                "*TOP NEWS* Metals",
            )
        ],
    )
    @pytest.mark.smoke
    def test_create_news_story_object_with_valid_story_id_and_get_data(
        self, open_session, id, story_title
    ):
        response = news.story.Definition(id).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_story_title(response, story_title)

    @allure.title(
        "Create News Story object with opened session and valid story ID - asynchronously"
    )
    @pytest.mark.caseid("34114895")
    @pytest.mark.parametrize(
        "id,story_title",
        [
            (
                "urn:newsml:reuters.com:20171030:nTOPMTL:169",
                "*TOP NEWS* Metals",
            )
        ],
    )
    async def test_create_news_story_object_with_valid_story_id_and_get_data_async(
        self, open_session_async, id, story_title
    ):
        response = await get_async_response_from_definition(news.story.Definition(id))
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_story_title(response, story_title)

    @allure.title(
        "Create News Story object using opened session with the invalid storyId "
    )
    @pytest.mark.caseid("34114895")
    @pytest.mark.parametrize("id", ["invalid"])
    def test_create_news_story_object_with_invalid_story_id(self, open_session, id):
        underlying_mark = open_session.type.name
        expected_error_message = ERROR_MESSAGE_PATTERN[underlying_mark]
        with pytest.raises(RDError) as error:
            news.story.Definition(id).get_data()
        assert str(error.value) == expected_error_message

    @allure.title("Create News Story object with closed session and check error")
    @pytest.mark.caseid("34114897")
    @pytest.mark.parametrize("id", ["urn:newsml:reuters.com:20201026:nPt6BSyBh"])
    async def test_create_news_story_object_with_closed_session(self, open_session_async, id):
        session = open_session_async
        await session.close_async()
        news_definition = news.story.Definition(id)
        with pytest.raises(ValueError) as error:
            await get_async_response_from_definition(news_definition)
        assert str(error.value) == "Session is not opened. Can't send any request"



# coding: utf8

from enum import Enum, unique


@unique
class IndexSpreadCompoundingMethod(Enum):
    ISDA_COMPOUNDING = "IsdaCompounding"
    ISDA_FLAT_COMPOUNDING = "IsdaFlatCompounding"
    NO_COMPOUNDING = "NoCompounding"



from typing import TYPE_CHECKING

from ._stream import (
    Events,
    FinalizedOrders,
    UniverseTypes,
    universe_type_arg_parser,
    finalized_orders_arg_parser,
    events_arg_parser,
)
from ._stream_facade import Stream
from ..._tools import create_repr, try_copy_to_list

if TYPE_CHECKING:
    from ..._types import OptStr, ExtendedParams, OptStrStrs
    from ..._core.session import Session


class Definition:
    """
    This class describes Analytics Trade Data Service.

    Parameters
    ----------
    universe : list, optional
        A list of RIC or symbol or user's id for retrieving trading analytics data.
    fields : list, optional
        A list of enumerate fields.
    events : Events, optional
        Enable/Disable the detail of order event in the streaming.
        Default: False
    finalized_orders : FinalizedOrders, optional
        Enable/Disable the cached of finalized order of current day in the streaming.
        Default: False
    filters : list, optional
        Set the condition of subset of trading streaming data.
    universe_type : UniverseTypes, optional
        A type of given universe can be RIC, Symbol or UserID.
        Default: UniverseTypes.RIC
    api: str, optional
        Specifies the data source. It can be updated/added using config file
    extended_params : dict, optional
        If necessary other parameters

    Methods
    -------
    get_stream(session=session)
        Get stream object of this definition

    Examples
    --------
    >>> from refinitiv.data.content import trade_data_service
    >>> definition = trade_data_service.Definition()
    """

    def __init__(
        self,
        universe: "OptStrStrs" = None,
        universe_type: UniverseTypes = UniverseTypes.UserID,
        fields: "OptStrStrs" = None,
        events: Events = Events.No,
        finalized_orders: FinalizedOrders = FinalizedOrders.No,
        filters: "OptStrStrs" = None,
        api: "OptStr" = None,
        extended_params: "ExtendedParams" = None,
    ):
        self._universe = try_copy_to_list(universe)
        self._universe_type = universe_type_arg_parser.get_str(universe_type)
        self._fields = try_copy_to_list(fields)
        self._events = events_arg_parser.get_str(events)
        self._finalized_orders = finalized_orders_arg_parser.get_str(finalized_orders)
        self._filters = try_copy_to_list(filters)
        self._api = api
        self._extended_params = extended_params

    def __repr__(self):
        return create_repr(
            self,
            middle_path="content.trade_data_service",
            content={"universe": self._universe},
        )

    def get_stream(self, session: "Session" = None) -> Stream:
        """
        Returns a streaming trading analytics subscription.

        Parameters
        ----------
        session : Session, optional
            The Session used by the TradeDataService to retrieve data from the platform

        Returns
        -------
        TradeDataStream

        Examples
        --------
        >>> from refinitiv.data.content import trade_data_service
        >>> definition = trade_data_service.Definition()
        >>> stream = definition.get_stream()
        >>> stream.open()
        """
        stream = Stream(
            session=session,
            universe=self._universe,
            universe_type=self._universe_type,
            fields=self._fields,
            events=self._events,
            finalized_orders=self._finalized_orders,
            filters=self._filters,
            api=self._api,
            extended_params=self._extended_params,
        )
        return stream



__all__ = (
    "search",
    "manage",
    "CrossCurrencyCurveDefinitionDescription",
    "CrossCurrencyCurveUpdateDefinition",
    "BidAskFieldsDescription",
    "BidAskFieldsFormulaDescription",
    "CrossCurrencyConstituentsDescription",
    "CrossCurrencyCurveParameters",
    "CrossCurrencyInstrumentDefinition",
    "CrossCurrencyInstrumentDescription",
    "CrossCurrencyInstrumentsSegment",
    "FieldDescription",
    "FieldFormulaDescription",
    "FormulaParameterDescription",
    "FxForwardInstrumentDefinition",
    "FxForwardInstrumentDescription",
    "FxForwardTurnFields",
    "FxSpotInstrumentDefinition",
    "FxSpotInstrumentDescription",
    "OverrideBidAsk",
    "OverrideBidAskFields",
    "OverrideFxForwardTurn",
    "TurnAdjustment",
    "InterpolationMode",
    "MainConstituentAssetClass",
    "QuotationMode",
    "RiskType",
    "StandardTurnPeriod",
)


from . import search, manage

from ...._curves._cross_currency_curves._definitions import (
    BidAskFieldsDescription,
    BidAskFieldsFormulaDescription,
    CrossCurrencyConstituentsDescription,
    CrossCurrencyCurveParameters,
    CrossCurrencyInstrumentDefinition,
    CrossCurrencyInstrumentDescription,
    CrossCurrencyInstrumentsSegment,
    FieldDescription,
    FieldFormulaDescription,
    FormulaParameterDescription,
    FxForwardInstrumentDefinition,
    FxForwardInstrumentDescription,
    FxForwardTurnFields,
    FxSpotInstrumentDefinition,
    FxSpotInstrumentDescription,
    OverrideBidAsk,
    OverrideBidAskFields,
    OverrideFxForwardTurn,
    TurnAdjustment,
)
from ...._curves._cross_currency_curves._definitions._create import (
    CrossCurrencyCurveDefinitionDescription,
)

from ...._curves._cross_currency_curves._definitions._update import (
    CrossCurrencyCurveUpdateDefinition,
)

from ._enums import (
    InterpolationMode,
    MainConstituentAssetClass,
    QuotationMode,
    RiskType,
    StandardTurnPeriod,
)



import allure
import numpy
import pytest

from refinitiv.data.content import custom_instruments as ci
from refinitiv.data.content.custom_instruments import SpreadAdjustmentMethod
from tests.integration.content.custom_instruments.conftest import (
    instrument_data,
    new_instrument_data,
    volume_based_udc,
    day_based_udc,
    manual_udc,
    new_day_based_udc,
)


@allure.suite("Content object - Custom Instrument")
@allure.feature("UDC Custom Instrument")
@allure.severity(allure.severity_level.CRITICAL)
class TestUDCCustomInstrument:
    @allure.title("Create UDC Custom Instrument")
    @pytest.mark.caseid("")
    @pytest.mark.parametrize(
        "udc",
        [volume_based_udc, day_based_udc, manual_udc],
        ids=["volume_based_udc", "day_based_udc", "manual_udc"],
    )
    def test_create_udc_custom_instrument(
        self,
        request,
        open_session_with_rdp_creds_for_ci,
        symbol,
        udc,
    ):
        cust_inst = ci.manage.create_udc(**instrument_data, symbol=symbol, udc=udc)

        for param, value in instrument_data.items():
            assert (
                cust_inst.__getattribute__(param) == value
            ), f"Attribute {param} was not created for {symbol}"
        assert (
            symbol in cust_inst.symbol
        ), f"Created symbol is not the same as requested"

        if "manual_udc" in request.node.callspec.id:
            for item in cust_inst.udc.rollover.manual_items:
                assert isinstance(item.year, int)
                assert isinstance(item.month, int)
                assert isinstance(item.start_date, numpy.datetime64)
            assert (
                cust_inst.udc.spread_adjustment.method
                == SpreadAdjustmentMethod.CLOSE_TO_CLOSE
            )
        else:
            assert (
                cust_inst.__getattribute__("udc") == udc
            ), "There is difference with UDC parameters"

        ci.manage.delete(symbol)

    @allure.title("Update UDC Custom Instrument objects with params")
    @pytest.mark.caseid("C42995815")
    @pytest.mark.parametrize(
        "udc,new_udc",
        [(day_based_udc, new_day_based_udc)],
    )
    def test_update_udc_custom_instrument(
        self, open_session_with_rdp_creds_for_ci, symbol, udc, new_udc
    ):
        cust_inst = ci.manage.create_udc(**instrument_data, symbol=symbol, udc=udc)

        for param, value in new_instrument_data.items():
            cust_inst.__setattr__(param, value)
        cust_inst.udc = new_udc
        cust_inst.save()

        updated_instrument = ci.manage.get(universe=symbol)

        for param, value in new_instrument_data.items():
            assert (
                updated_instrument.__getattribute__(param) == value
            ), f"Attribute {param} was not updated  for {symbol}"
        assert cust_inst.udc == new_udc, f"UDC object isn't updated"
        ci.manage.delete(symbol)



"""Base for templates namespaces"""

from collections import deque


class Namespace:
    def __init__(self, target=None, **kwargs):
        # Don't have an idea yet how to it better without need to pass TargetTemplate
        # to Namespace each time or creating circular dependencies
        if target is None:
            from .base import TargetTemplate

            self.target = TargetTemplate
        for key, value in kwargs.items():
            if "." in key:
                raise ValueError(
                    f"Name in namespace can't contain dot. Value with dot encountered: {value}"
                )
            if not isinstance(value, (Namespace, self.target)):
                raise TypeError(
                    f"Value in namespace must be either SearchTemplate or other Namespace. {type(value)} encountered"
                )

        self._data = kwargs

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, name: str):
        return self._data.__getitem__(name)

    def get(self, path: str):
        """Get namespace or key by path"""
        keys = deque(path.split("."))
        result = self
        while keys:
            result = result[keys.popleft()]
            if isinstance(result, self.target) and keys:
                return None
        return result

    def keys(self):
        return self._data.keys()



from typing import TYPE_CHECKING

from ..._content_data import Data
from ..._content_provider_layer import ContentUsageLoggerMixin
from ...._content_type import ContentType
from ...._tools import validate_bool_value, try_copy_to_list
from ....delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ...._types import ExtendedParams, StrStrings


class Definition(
    ContentUsageLoggerMixin[BaseResponse[Data]],
    DataProviderLayer[BaseResponse[Data]],
):
    """
    This class describe parameters to retrieves estimates actuals values for KPI Measures for reported interim periods.

    Parameters
    ----------
    universe: str, list of str
        The Universe parameter allows the user to define the companies for which the content is returned.

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params: ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import estimates
    >>> definition = estimates.view_actuals_kpi.interim.Definition(universe="BNPP.PA")
    >>> response = definition.get_data()
    """

    _USAGE_CLS_NAME = "Estimates.ActualsKPI.InterimDefinition"

    def __init__(
        self,
        universe: "StrStrings",
        use_field_names_in_headers: bool = False,
        extended_params: "ExtendedParams" = None,
    ):
        validate_bool_value(use_field_names_in_headers)
        universe = try_copy_to_list(universe)

        super().__init__(
            ContentType.ESTIMATES_VIEW_ACTUALS_KPI_INTERIM,
            universe=universe,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



from typing import TYPE_CHECKING

from ...delivery._data._data_provider import (
    ContentValidator,
)

if TYPE_CHECKING:
    from ...delivery._data._data_provider import ParsedData


class IPAContentValidator(ContentValidator):
    _NAME_DATA = "data"

    @classmethod
    def content_data_is_not_none(cls, data: "ParsedData") -> bool:
        if data.content_data.get(cls._NAME_DATA) is None:
            data.error_codes = 1
            data.error_messages = "Content data is None"
            return False

        return True

    @classmethod
    def any_element_have_no_error(cls, data: "ParsedData") -> bool:
        elements = data.content_data.get(cls._NAME_DATA)
        if isinstance(elements, list):
            counter = len(elements) or 1
            for element in elements:
                if not hasattr(element, "get"):
                    counter -= 1
                    data.error_messages = (
                        f"Invalid data type={type(element)}, data={element}"
                    )
                    continue

                error = element.get("error")

                if error:
                    counter -= 1
                    error_code = error.get("code")
                    data.error_codes.append(error_code)
                    error_message = error.get("message")
                    data.error_messages.append(error_message)

            if counter == 0:
                return False

        return True



from ._discovery_universe import DiscoveryUniverse


class Screener(DiscoveryUniverse):
    """
    Class to get data from screen function.

    Parameters
    ----------
    expression : str
        screen expression


    Examples
    --------
    >>> screener = Screener('U(IN(Equity(active,public,primary))/*UNV:Public*/), IN(TR.HQCountryCode,"AR"), IN(TR.GICSIndustryCode,"401010")')
    >>> print(list(screener))
    """

    def __init__(self, expression):
        super().__init__(f"screen({expression})")



import pytest

from refinitiv.data.delivery.cfs._data_types import BucketData
from refinitiv.data.delivery.cfs._iter_object import BaseCFSObject


@pytest.mark.parametrize(
    "input_session, expected_session",
    [
        (None, None),
        ([], []),
        ("filename", "filename"),
    ],
)
def test_attribute_session_in_BaseCFSObject(input_session, expected_session):
    # when
    obj = BaseCFSObject({}, BucketData, session=input_session)

    # then
    assert hasattr(obj, "_session")
    assert obj._session == expected_session



from ....._object_definition import ObjectDefinition


class DeleteRequest(ObjectDefinition):
    """
    Delete a cross currency curve definition

    Parameters
    ----------
    id : str
        The identifier of the cross currency definition.
    """

    def __init__(self, id: str) -> None:
        super().__init__()
        self.id = id

    @property
    def id(self):
        """
        :return: str
        """
        return self._get_parameter("id")

    @id.setter
    def id(self, value):
        self._set_parameter("id", value)



import logging
from types import SimpleNamespace

import pytest

from refinitiv.data.content.esg.bulk._actions import Actions
from refinitiv.data.content.esg.bulk._db_manager import create_db_manager
from refinitiv.data.content.esg.bulk._file_manager import FileManager


def create_actions():
    actions = SimpleNamespace()
    actions.add = lambda *_, **__: None
    actions.updated = lambda *_, **__: None
    actions.update = lambda *_, **__: None
    actions.get_created_tables = lambda: []
    return actions


logger = logging.getLogger("test-db-manager")

config = {
    "connection.parameters.database": ":memory:",
    "create-table-queries": [],
    "insert-queries": [],
    "connection.module": "sqlite3",
}

id_counter = 0


@pytest.fixture(scope="function")
def actions():
    global id_counter
    id_counter += 1
    actions_ = Actions(id_=id_counter, logpath="./tests/unit/content/esg/bulk/log.txt")
    yield actions_
    actions_.dispose()


@pytest.fixture
def file_manager(actions):
    file_manager = FileManager(
        logger=logger,
        actions=actions,
        package_name="",
        bucket="",
        path="",
        session=None,
    )
    return file_manager


@pytest.fixture()
def db_manager():
    manager = create_db_manager(config, create_actions(), logger)
    return manager


def raise_(exception):
    raise exception



import json

import allure
import pytest

from refinitiv.data.content.ipa.curves import zc_curve_definitions
from tests.integration.content.ipa.curves.conftest import on_response
from tests.integration.content.ipa.curves.zc_curve_definition.conftest import (
    check_sc_curve_definitions_response,
    zc_curve_definition_01,
    zc_curve_definition_02, zc_curve_definition_universe,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - ZcCurveDefinition")
@allure.feature("Content object - ZcCurveDefinition")
@allure.severity(allure.severity_level.CRITICAL)
class TestZcCurveDefinition:
    @allure.title("Create a zc_curve_definition object with valid params and get data")
    @pytest.mark.caseid("39597176")
    @pytest.mark.smoke
    def test_create_zc_curve_definition_with_valid_params(self, open_session):
        response = zc_curve_definition_01().get_data(
            on_response=lambda response, definition, session: on_response(
                response, definition, open_session
            )
        )

        check_sc_curve_definitions_response(response=response, expected_currency="EUR")
        check_dataframe_column_date_for_datetime_type(response)


    @allure.title("Create zc curve definition object with invalid source and get data")
    @pytest.mark.caseid("39597177")
    def test_create_empty_zc_curve_definition(self, open_session):
        response = zc_curve_definitions.Definition(source="INVALID_SOURCE").get_data()

        assert response.data.df.empty

    @allure.title("Create a zc curve definition object async")
    @pytest.mark.caseid("39597178")
    async def test_create_zc_curve_definition_async(self, open_session_async):
        response_01, response_02 = await get_async_response_from_definitions(
            zc_curve_definition_01(), zc_curve_definition_02()
        )

        check_sc_curve_definitions_response(
            response=response_01, expected_currency="EUR"
        )

    @allure.title("Create a zc_curve definition with extended parameters")
    @pytest.mark.caseid("C40214933")
    def test_get_zc_curve_definition_with_extended_params(self, open_session):
        response = zc_curve_definitions.Definition(
            source="Refinitiv",
            extended_params={
                "indexName": "EURIBOR",
                "source": "Refinitiv",
                "valuationDate": "2020-07-01",
            },
        ).get_data()

        check_sc_curve_definitions_response(response=response, expected_currency="EUR")

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == zc_curve_definition_universe
        ), f"Extended params are applied incorrectly"
        assert response.data.df.indexName[0] == "EURIBOR"






from typing import Union

import numpy as np
from scipy import interpolate

from ._surface_point import SurfacePoint
from .._enums import Axis
from ..._content_provider import value_arg_parser, x_arg_parser, y_arg_parser
from ..._curves._models._curve import Curve
from ....._tools._common import cached_property


class Surface:
    """
    The actual volatility surface

    Parameters
    -------
    row : np.array
        Specifies a data of discrete values for the y-axis
    column : np.array
        Specifies a data of discrete values for the x-axis
    matrix : np.array
        Specifies whether the calculated volatilities a matrix.

    Methods
    -------
        get_curve(value, axis): Curve

        get_point(x, y): SurfacePoint

    Examples
    -------
    >>> from refinitiv.data.content.ipa.surfaces import eti
    >>> definition = eti.Definition(
    ...     underlying_definition=eti.EtiSurfaceDefinition(
    ...         instrument_code="BNPP.PA@RIC"
    ...     ),
    ...     surface_parameters=eti.EtiCalculationParams(
    ...         price_side=eti.PriceSide.MID,
    ...         volatility_model=eti.VolatilityModel.SVI,
    ...         x_axis=eti.Axis.STRIKE,
    ...         y_axis=eti.Axis.DATE,
    ...     ),
    ...     surface_tag="1_MATRIX",
    ...     surface_layout=eti.SurfaceLayout(
    ...         format=eti.Format.MATRIX,
    ...         y_point_count=10
    ...     ),
    ... )
    >>> response = definition.get_data()
    >>> surface = response.data.surface
    >>> curve = surface.get_curve("25.35", eti.Axis.X)
    >>> curve = surface.get_curve("2021-08-20", eti.Axis.Y)
    >>> point = surface.get_point(25.35, "2020-08-20")

    """

    def __init__(self, row: np.array, column: np.array, matrix: np.array) -> None:
        self._row = row
        self._column = column
        self._matrix = matrix

    @cached_property
    def _values_by_axis(self):
        return {
            Axis.X: self.x,
            Axis.Y: self.y,
            Axis.Z: self.z,
        }

    @cached_property
    def _interp(self):
        """
        For interpolate.interp2d axes are:

           |--→ column=x
           ↓
         row=y

        """
        return interpolate.interp2d(
            x=self._column.astype(float),
            y=self._row.astype(float),
            z=self._matrix.astype(float),
        )

    @property
    def x(self):
        return self._row

    @property
    def y(self):
        return self._column

    @property
    def z(self):
        return self._matrix

    def get_curve(
        self,
        value: Union[str, float, int, np.datetime64],
        axis: Axis,
    ) -> Curve:
        """
        Get curve

        Parameters
        ----------
        value: str, float, int, np.datetime64
            Establishment value
        axis: Axis
            The enumerate that specifies the unit for the axis

        Returns
        -------
        Curve
        """
        values = self._values_by_axis.get(axis, [])

        value = value_arg_parser.parse(value)

        axis_x, axis_y = [], []

        is_axis_x = axis is Axis.X
        is_axis_y = axis is Axis.Y
        is_axis_z = axis is Axis.Z

        # interpolate
        if value not in values and not is_axis_z:

            value = value_arg_parser.get_float(value)

            if is_axis_x:
                axis_x = self._column
                axis_y = self._interp(
                    x=self._column.astype(float),
                    y=value,
                )

            elif is_axis_y:
                axis_x = self._interp(
                    x=value,
                    y=self._row.astype(float),
                )
                axis_x = axis_x[:, 0]
                axis_y = self._row

        else:
            index = np.where(values == value)

            if is_axis_x:
                axis_x = self.z[index, :]
                axis_x = axis_x[0][0]
                axis_y = self.y

            elif is_axis_y:
                axis_x = self.x
                axis_y = self.z[:, index]
                axis_y = axis_y[:, 0][:, 0]

            elif is_axis_z:
                axis_x = self.x
                axis_y = self.y

        return Curve(axis_x, axis_y)

    def get_point(
        self,
        x: Union[str, int, float, np.datetime64],
        y: Union[str, int, float, np.datetime64],
    ) -> SurfacePoint:
        """
        Get SurfacePoint

        Parameters
        ----------
        x: str, float, int, np.datetime64
            X axis
        y: str, float, int, np.datetime64
            Y axis

        Returns
        -------
        SurfacePoint
        """

        x = x_arg_parser.parse(x)
        y = y_arg_parser.parse(y)

        if x in self.x and y in self.y:
            (row,) = np.where(self._row == x)
            (column,) = np.where(self._column == y)
            z = self._matrix[row, column]
            z = z[0]

        # interpolate
        else:
            x_as_float = value_arg_parser.get_float(x)
            y_as_float = value_arg_parser.get_float(y)
            z = self._interp(x=y_as_float, y=x_as_float)
            z = z[0]

        return SurfacePoint(x, y, z)



import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data.content.ipa._curves._curves_data_provider import (
    get_curves_maker,
    CurvesResponseFactory,
    forward_curve_build_df,
)
from refinitiv.data.delivery._data._data_provider import ParsedData
from tests.unit.content.ipa._curves.conftest import (
    zc_curves_json,
    forward_curves_json,
    zc_curve_definitions_json,
)


@pytest.mark.parametrize(
    "input_value", [ContentType.FORWARD_CURVE, ContentType.ZC_CURVES]
)
def test_get_curves_maker(input_value):
    # when
    testing_value = get_curves_maker(input_value)

    # then
    assert testing_value is not None


def test_get_curves_maker_will_raise_exception_if_invalid_value():
    # given
    input_value = None

    # then
    with pytest.raises(ValueError):
        # when
        get_curves_maker(input_value)


def test_create_forward_curves():
    # given
    create_forward_curves = get_curves_maker(ContentType.FORWARD_CURVE)

    # when
    curves = create_forward_curves(forward_curves_json)

    # then
    assert len(curves) > 1


def test_forward_curve_build_df():
    # given
    input_raw = {
        "data": [
            {
                "error": {
                    "id": "b6f9797d-72c8-4baa-84eb-6a079fc40ec5/b6f9797d-72c8-4baa-84eb-6a079fc40ec5",
                    "code": "QPS-Curves.6",
                    "message": "Invalid input: curveDefinition is missing",
                }
            },
            {
                "curveTag": "test_curve",
                "curveParameters": {
                    "interestCalculationMethod": "Dcb_Actual_Actual",
                    "priceSide": "Mid",
                    "calendarAdjustment": "Calendar",
                    "calendars": ["EMU_FI"],
                    "compoundingType": "Compounded",
                    "useConvexityAdjustment": True,
                    "useSteps": False,
                    "valuationDate": "2022-02-09",
                },
                "curveDefinition": {
                    "availableTenors": ["OIS", "1M", "3M", "6M", "1Y"],
                    "availableDiscountingTenors": ["OIS", "1M", "3M", "6M", "1Y"],
                    "currency": "EUR",
                    "mainConstituentAssetClass": "Swap",
                    "riskType": "InterestRate",
                    "indexName": "EURIBOR",
                    "source": "Refinitiv",
                    "name": "EUR EURIBOR Swap ZC Curve",
                    "id": "9d619112-9ab3-45c9-b83c-eb04cbec382e",
                    "discountingTenor": "OIS",
                    "ignoreExistingDefinition": False,
                    "owner": "Refinitiv",
                },
                "forwardCurves": [
                    {
                        "curvePoints": [
                            {
                                "endDate": "2021-02-01",
                                "startDate": "2021-02-01",
                                "discountFactor": 1.0,
                                "ratePercent": 7.040811073443143,
                                "tenor": "0D",
                            },
                            {
                                "endDate": "2021-02-04",
                                "startDate": "2021-02-01",
                                "discountFactor": 0.999442450671571,
                                "ratePercent": 7.040811073443143,
                                "tenor": "1D",
                            },
                        ],
                        "forwardCurveTag": "ForwardTag",
                        "forwardStart": "2021-02-01",
                        "indexTenor": "3M",
                    }
                ],
            },
        ]
    }

    expected_str = (
        "     endDate  startDate  discountFactor  ratePercent tenor\n"
        "0 2021-02-01 2021-02-01             1.0     7.040811    0D\n"
        "1 2021-02-04 2021-02-01        0.999442     7.040811    1D"
    )

    # when
    testing_df = forward_curve_build_df(input_raw)

    # then
    assert testing_df.to_string() == expected_str


def test_create_zc_curves():
    # given
    create_zc_curves = get_curves_maker(ContentType.ZC_CURVES)

    # when
    curves = create_zc_curves(zc_curves_json)

    # then
    assert len(curves) > 1


def test_curves_response_factory_create_response_from_one_curve():
    # given
    response_factory = CurvesResponseFactory()
    data = ParsedData({}, {}, zc_curves_json)
    kwargs = {
        "universe": object(),
        "__content_type__": ContentType.ZC_CURVES,
    }

    # when
    response = response_factory.create_success(data, **kwargs)

    # then
    assert response.data.curve is not None


def test_curves_response_factory_create_response_from_many_curves():
    # given
    response_factory = CurvesResponseFactory()
    data = ParsedData({}, {}, forward_curves_json)
    kwargs = {
        "universe": [object(), object()],
        "__content_type__": ContentType.FORWARD_CURVE,
    }

    # when
    response = response_factory.create_success(data, **kwargs)

    # then
    assert len(response.data.curves) == 2


def test_curves_response_factory_create_response_with_data_with_df():
    # given
    response_factory = CurvesResponseFactory()
    data = ParsedData({}, {}, zc_curves_json)
    kwargs = {
        "universe": object(),
        "__content_type__": ContentType.ZC_CURVES,
    }

    # when
    response = response_factory.create_success(data, **kwargs)

    # then
    assert response.data.df.empty is False


def test_curves_response_factory_create_response_for_zc_curve_definitions_with_data_with_df():
    # given
    response_factory = CurvesResponseFactory()
    data = ParsedData({}, {}, zc_curve_definitions_json)
    kwargs = {
        "universe": object(),
        "__content_type__": ContentType.ZC_CURVE_DEFINITIONS,
    }

    # when
    response = response_factory.create_success(data, **kwargs)

    # then
    assert "currency" in response.data.df


def test_curves_response_factory_create_response_for_zc_curve_definitions_with_data_with_empty_df():
    # given
    response_factory = CurvesResponseFactory()
    data = ParsedData({}, {}, {"data": []})
    kwargs = {
        "universe": object(),
        "__content_type__": ContentType.ZC_CURVE_DEFINITIONS,
    }

    # when
    response = response_factory.create_success(data, **kwargs)

    # then
    assert response.data.df.empty is True



from refinitiv.data.content.estimates.view_actuals_kpi import annual, interim


def test_annual(create_estimates_kpi_definition):
    annual.Definition(**create_estimates_kpi_definition)


def test_interim(create_estimates_kpi_definition):
    interim.Definition(**create_estimates_kpi_definition)


def test_annual_error(create_estimates_kpi_definition_with_error):
    try:
        annual.Definition(**create_estimates_kpi_definition_with_error)
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)


def test_interim_error(create_estimates_kpi_definition_with_error):
    try:
        interim.Definition(**create_estimates_kpi_definition_with_error)
    except (TypeError, ValueError):
        assert True
    except Exception as e:
        assert False, str(e)



from typing import Union, TYPE_CHECKING

from ._response import ContribResponse, NullContribResponse
from ...._core.session import get_valid_session
from ...._tools import make_callback

if TYPE_CHECKING:
    from ._type import ContribType
    from ...._core.session import Session
    from ...._types import OptStr, OptCall


def contribute(
    name: str,
    fields: dict,
    *,
    service: "OptStr" = None,
    contrib_type: Union[str, "ContribType", None] = None,
    session: "Session" = None,
    api: "OptStr" = None,
    on_ack: "OptCall" = None,
    on_error: "OptCall" = None,
) -> "ContribResponse":
    """
    Function to send OffStream contribution request.

    Parameters
    ----------
    name: string
        RIC to retrieve instrument stream.

    fields: dict{field:value}
        Specify fields and values to contribute.

    service: string, optional
        Specify the service to contribute on.
        Default: None

    contrib_type: Union[str, ContribType], optional
        Define the contribution type : ["Refresh", "Update"]
        Default: "Update"

    session: Session, optional
        Specify the session used to contribute

    api: string, optional
        specific name of contrib streaming defined in config file.
        i.e. 'streaming.contrib.endpoints.my_server'
        Default: 'streaming.contrib.endpoints.main'

    on_ack : function, optional
        Callback function for on_ack event to check contribution result

    on_error : function, optional
        Callback function for on_error event

    Returns
    ----------
    ContribResponse

    Examples
    --------
    Prerequisite: The contrib_session must be opened
    >>> import refinitiv.data as rd
    >>> def on_ack_callback(ack_msg, stream):
    ...     print("Receive Ack response:", ack_msg)
    >>> def on_error_callback(error_msg, stream):
    ...     print("Receive Error:", error_msg)
    >>> update = {
    ...     "ASK": 1.23,
    ...     "BID": 1.24
    ... }
    >>> response = rd.delivery.omm_stream.contribute(
    ...     name="EUR=",
    ...     fields=update,
    ...     service="SVC_CONTRIB",
    ...     on_ack=on_ack_callback,
    ...     on_error=on_error_callback
    ... )
    """
    from ..._stream._stream_factory import create_offstream_contrib

    session = get_valid_session(session)
    offstream = create_offstream_contrib(
        session=session,
        name=name,
        api=api,
        domain="MarketPrice",
        service=service,
    )
    on_ack and offstream.on_ack(make_callback(on_ack))
    on_error and offstream.on_error(make_callback(on_error))
    try:
        offstream.open()
    except ConnectionError:
        response = NullContribResponse()
        on_error and on_error(offstream.get_contrib_error_message(), offstream)
    else:
        response = offstream.contribute(fields, contrib_type)
        offstream.close()
    return response


async def contribute_async(
    name: str,
    fields: dict,
    *,
    service: "OptStr" = None,
    contrib_type: Union[str, "ContribType", None] = None,
    session: "Session" = None,
    api: "OptStr" = None,
    on_ack: "OptCall" = None,
    on_error: "OptCall" = None,
) -> "ContribResponse":
    """
    Function to send asynchrnous OffStream contribution request.

    Parameters
    ----------
    name: string
        RIC to retrieve instrument stream.

    fields: dict{field:value}
        Specify fields and values to contribute.

    service: string, optional
        Specify the service to contribute on.
        Default: None

    contrib_type: Union[str, ContribType], optional
        Define the contribution type
        Default: "Update"

    session: Session, optional
        Specify the session used to contribute

    api: string, optional
        specific name of contrib streaming defined in config file.
        i.e. 'streaming.contrib.endpoints.my_server'
        Default: 'streaming.contrib.endpoints.main'

    on_ack : function, optional
        Callback function for on_ack event to check contribution result

    on_error : function, optional
        Callback function for on_error event

    Returns
    ----------
    ContribResponse

    Examples
    --------
    Prerequisite: The contrib_session must be opened.
    >>> import refinitiv.data as rd
    >>> def on_ack_callback(ack_msg, stream):
    ...     print("Receive Ack response:", ack_msg)
    >>> def on_error_callback(error_msg, stream):
    ...     print("Receive Error:", error_msg)
    >>> update = {
    ...     "ASK": 1.23,
    ...     "BID": 1.24
    ... }
    >>> response = await rd.delivery.omm_stream.contribute_async(
    ...     "EUR=",
    ...     fields=update,
    ...     service="SVC_CONTRIB",
    ...     on_ack=on_ack_callback,
    ...     on_error=on_error_callback
    ... )
    """
    from ..._stream._stream_factory import create_offstream_contrib

    session = get_valid_session(session)
    offstream = create_offstream_contrib(
        session=session,
        name=name,
        api=api,
        domain="MarketPrice",
        service=service,
    )
    on_ack and offstream.on_ack(make_callback(on_ack))
    on_error and offstream.on_error(make_callback(on_error))
    try:
        await offstream.open_async()
    except ConnectionError:
        response = NullContribResponse()
        on_error and on_error(offstream.get_contrib_error_message(), offstream)
    else:
        response = await offstream.contribute_async(fields, contrib_type)
        offstream.close()
    return response



import pytest

from refinitiv.data.content.pricing._stream_facade import Stream
from tests.unit.conftest import StubSession


@pytest.fixture(scope="function")
def pricing_stream():
    session = StubSession(is_open=True)
    pricing_stream = Stream(session=session)
    return pricing_stream



# coding: utf8

from enum import Enum, unique


@unique
class FxBinaryType(Enum):
    DIGITAL = "Digital"
    NO_TOUCH = "NoTouch"
    NONE = "None"
    ONE_TOUCH_DEFERRED = "OneTouchDeferred"
    ONE_TOUCH_IMMEDIATE = "OneTouchImmediate"



import pandas as pd
import pytest

from refinitiv.data.content.ipa.financial_contracts._contracts_data_provider import (
    Data,
    financial_contracts_build_df,
)


@pytest.mark.parametrize(
    "input_data,input_type,expected_data",
    [
        ([None], "Date", pd.NaT),
        (["2032-02-28"], "Date", pd.to_datetime("2032-02-28")),
        (["2032-02-28T00:00:00Z"], "DateTime", pd.to_datetime("2032-02-28")),
    ],
)
def test_build_df_with_datetime(input_data, input_type, expected_data):
    # given
    expected_df = pd.DataFrame({"IssueDate": {0: expected_data}})
    input_raw = {
        "headers": [
            {"type": input_type, "name": "IssueDate"},
        ],
        "data": [input_data],
    }

    # then
    testing = Data(raw=input_raw, dfbuilder=financial_contracts_build_df)

    # when
    try:
        diff_df = testing.df.compare(expected_df)
    except ValueError as e:
        assert False, str(e)
    else:
        assert diff_df.empty is True, diff_df.to_string()



from pandas import DataFrame

from ._tools import _get_query_parameter
from .._data._endpoint_data import EndpointData


class BaseData(EndpointData):
    def __init__(self, raw, iter_object=None, **kwargs):
        EndpointData.__init__(self, raw, **kwargs)
        self._raw["skip_token"] = _get_query_parameter(
            "skipToken", self._raw.get("@nextLink", None)
        )
        self._iter_object = iter_object
        self._dataframe = None

    @property
    def df(self):
        if self._dataframe is None:
            value = self.raw.get("value") or [self.raw]
            columns = set()
            for i in value:
                columns = columns | i.keys()
            columns = tuple(columns)
            data = [
                [value[key] if key in value else None for key in columns]
                for value in value
            ]
            self._dataframe = DataFrame(data, columns=columns)

        return self._dataframe


class BucketData(BaseData):
    @property
    def buckets(self):
        return self._iter_object


class FileSetData(BaseData):
    @property
    def file_sets(self):
        return self._iter_object


class PackageData(BaseData):
    @property
    def packages(self):
        return self._iter_object


class FileData(BaseData):
    @property
    def files(self):
        return self._iter_object



from typing import TYPE_CHECKING, Union, Iterable, Tuple

from pandas import DataFrame

from ._containers import (
    FieldsContainer,
    UniverseContainer,
    ADCDataContainer,
    HPAndCustInstDataContainer,
)
from ._data_provider import DataProvider, convert_types
from .context_collection import get_adc_context
from .context_collection._context_factory import get_hp_and_custinst_context
from .._core.session import get_default, raise_if_closed
from .._errors import ScopeError
from .._tools import (
    DEBUG,
)
from .._tools._dataframe import (
    convert_dtypes,
)
from ..content import fundamental_and_reference
from ..content.fundamental_and_reference._data_grid_type import (
    get_data_grid_type_by_session,
)
from ..errors import RDError
from ..usage_collection._filter_types import FilterType
from ..usage_collection._logger import get_usage_logger
from ..usage_collection._utils import ModuleName

if TYPE_CHECKING:
    from logging import Logger


def get_log_string(fields, universe):
    return f"Fields: {fields} for {universe}"


def show_requests_in_log(logger: "Logger", response: "Response", universe, fields):
    request_messages = response.request_message
    statuses = response.http_status
    if not isinstance(response.request_message, list):
        request_messages = [response.request_message]
        statuses = [response.http_status]

    for request, status in zip(request_messages, statuses):
        path = request.url.path
        cur_universe = path.rsplit("/", 1)[-1]
        if cur_universe not in universe:
            cur_universe = universe
        logger.info(
            f"Request to {path} with {get_log_string(fields, cur_universe)}\n"
            f"status: {status}\n"
        )


def get_adc_data(
    params: dict, logger: "Logger", raise_if_error=False
) -> Tuple[dict, "DataFrame"]:
    fields = params.get("fields", "")
    universe = params["universe"]
    logger.info(f"Requesting {get_log_string(fields, universe)} \n")
    definition = fundamental_and_reference.Definition(**params)
    raw = {}
    df = DataFrame()
    try:
        response = definition.get_data()
        raw = response.data.raw
        df = response.data.df
    except ScopeError as e:
        if raise_if_error:
            raise e
        logger.error(str(e))
    except RDError as e:
        if raise_if_error:
            raise e
        logger.debug(f"Failure response into content layer: {str(e)}")
    except Exception as e:
        if DEBUG:
            raise e
        if raise_if_error:
            raise e
        logger.exception(f"Failure sending request with {definition}")
    else:
        show_requests_in_log(logger, response, universe, fields)
    return raw, df


def get_data_from_stream(universe, fields, logger):
    from . import Stream

    logger.info(f"Requesting pricing info for fields={fields} via websocket\n")
    stream = Stream(universe=universe, fields=fields)
    columns, data, df = None, None, DataFrame()

    try:
        stream.open(with_updates=False)
        columns, data = get_columns_and_data_from_stream(stream, fields)
        df = stream.get_snapshot(fields=fields)
        if len(df.columns) == 1 or not any([_stream.fields for _stream in stream]):
            df = DataFrame()
        else:
            df = convert_dtypes(df)
        stream.close()

    except Exception as e:
        logger.debug(f"Failure retrieving data for {stream._stream.universe}")

    return columns, data, df


def get_columns_from_stream(stream):
    columns = set()
    for _stream in stream:
        fields = _stream.fields or []
        columns.update(fields)
    return list(columns)


def get_columns_and_data_from_stream(stream, fields):
    stream_columns = get_columns_from_stream(stream)
    if fields:
        columns = [i for i in fields if i in stream_columns]
    else:
        columns = stream_columns
    data = {
        _stream.name: convert_types([_stream[column] for column in columns], columns)
        for _stream in stream
    }
    return columns, data


def update_universe(raw, _universe):
    index = 0  # instrument
    data = raw.get("data")
    if data and all(isinstance(i[index], str) for i in data):
        universe = [i[index] for i in data]
    else:
        universe = _universe
    return universe


def get_data(
    universe: Union[str, Iterable[str]],
    fields: Union[str, Iterable[str], None] = None,
    parameters: Union[str, dict, None] = None,
    use_field_names_in_headers: bool = False,
) -> DataFrame:
    """
    Retrieves pricing snapshots, as well as Fundamental and Reference data.

    Parameters
    ----------
    universe: str | list
        Instruments to request
    fields: str | list, optional
        Fields to request
    parameters: str | dict, optional
        Single key=value global parameter or dictionary of global parameters to request
    use_field_names_in_headers: bool, default False
        If True - returns field name as column headers for data instead of title

    Returns
    -------
    pandas.DataFrame

    Examples
    --------
    >>> get_data(universe=['IBM.N', 'VOD.L'], fields=['BID', 'ASK'])
    >>> get_data(
    ...     universe=['GOOG.O', 'AAPL.O'],
    ...     fields=['TR.EV','TR.EVToSales'],
    ...     parameters = {'SDate': '0CY', 'Curn': 'CAD'}
    ...)
    """
    session = get_default()
    raise_if_closed(session)

    logger = session.logger()

    # Library usage logging
    get_usage_logger().log_func(
        name=f"{ModuleName.ACCESS}.get_data",
        func_path=f"{__name__}.get_data",
        kwargs=dict(
            universe=universe,
            fields=fields,
            parameters=parameters,
            use_field_names_in_headers=use_field_names_in_headers,
        ),
        desc={FilterType.SYNC, FilterType.LAYER_ACCESS},
    )

    universe = UniverseContainer(universe)
    fields = FieldsContainer(fields)
    data_grid_type = get_data_grid_type_by_session(session)
    adc = get_adc_context(data_grid_type, universe, fields)
    hp_and_cust_inst = get_hp_and_custinst_context(data_grid_type, universe, fields)

    adc_raw, adc_df = None, None
    if adc.can_get_data:
        adc_raw, adc_df = get_adc_data(
            params={
                "universe": universe.adc,
                "fields": fields.adc,
                "parameters": parameters,
                "use_field_names_in_headers": use_field_names_in_headers,
            },
            logger=logger,
        )

    adc_data = ADCDataContainer(adc_raw, adc_df, fields)
    universe.calc_hp(adc_raw)

    stream_columns, stream_data, stream_df = None, None, None
    if hp_and_cust_inst.can_get_data:
        stream_columns, stream_data, stream_df = get_data_from_stream(
            universe.hp_and_cust_inst, fields.hp, logger
        )

    hp_and_cust_inst_data = HPAndCustInstDataContainer(
        stream_columns, stream_data, stream_df
    )

    adc.set_data(adc_data, hp_and_cust_inst_data)
    hp_and_cust_inst.set_data(adc_data, hp_and_cust_inst_data)

    provider = DataProvider()
    return provider.get_df(
        adc,
        hp_and_cust_inst,
        adc_data,
        hp_and_cust_inst_data,
        use_field_names_in_headers,
    )



from typing import TYPE_CHECKING, Optional

from ..._content_data import Data
from ..._content_provider_layer import ContentUsageLoggerMixin
from ...._content_type import ContentType
from ...._tools import validate_types, validate_bool_value, try_copy_to_list
from ....delivery._data._data_provider import DataProviderLayer, BaseResponse

if TYPE_CHECKING:
    from ...._types import ExtendedParams, StrStrings


class Definition(
    ContentUsageLoggerMixin[BaseResponse[Data]],
    DataProviderLayer[BaseResponse[Data]],
):
    """
    This class describe parameters to retrieve company ownership details. Also, information on
    the top 20 fund shareholders invested in the requested company.

    Parameters
    ----------
    universe: str, list of str
        The Universe parameter allows the user to define the companies for which the content is returned.

    limit: int, optional
        The limit parameter is used for paging. It allows users to select the number of records to be returned.
        Default page size is 100 or 20 (depending on the operation).

    use_field_names_in_headers: bool, optional
        Return field name as column headers for data instead of title

    extended_params : ExtendedParams, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content import ownership
    >>> definition = ownership.fund.investors.Definition("TRI.N")
    >>> response = definition.get_data()
    """

    _USAGE_CLS_NAME = "Ownership.Fund.InvestorsDefinition"

    def __init__(
        self,
        universe: "StrStrings",
        limit: Optional[int] = None,
        use_field_names_in_headers: bool = False,
        extended_params: "ExtendedParams" = None,
    ):
        validate_types(limit, [int, type(None)], "limit")
        validate_bool_value(use_field_names_in_headers)
        universe = try_copy_to_list(universe)

        super().__init__(
            ContentType.OWNERSHIP_FUND_INVESTORS,
            universe=universe,
            limit=limit,
            use_field_names_in_headers=use_field_names_in_headers,
            extended_params=extended_params,
        )



from enum import unique, Enum


@unique
class OpenState(Enum):
    """
    Session lifecycle state.
        Opened : the session is open and ready to work.
        Pending : the session is waiting to be opened or closed.
        Closed : the session is closed.
    """

    Opened = "Opened"
    Pending = "Pending"
    Closed = "Closed"



# coding: utf8

from typing import Optional, TYPE_CHECKING, List

from ._swap_definition import SwapInstrumentDefinition
from .._base_definition import BaseDefinition
from ....._tools import try_copy_to_list

if TYPE_CHECKING:
    from ....._types import ExtendedParams, OptStrStrs
    from . import LegDefinition, PricingParameters


class Definition(BaseDefinition):
    def __init__(
        self,
        instrument_tag: Optional[str] = None,
        instrument_code: Optional[str] = None,
        trade_date: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        tenor: Optional[str] = None,
        legs: Optional[List["LegDefinition"]] = None,
        is_non_deliverable: Optional[bool] = None,
        settlement_ccy: Optional[str] = None,
        start_tenor: Optional[str] = None,
        template: Optional[str] = None,
        fields: "OptStrStrs" = None,
        pricing_parameters: Optional["PricingParameters"] = None,
        extended_params: "ExtendedParams" = None,
    ):
        legs = try_copy_to_list(legs)
        fields = try_copy_to_list(fields)
        definition = SwapInstrumentDefinition(
            legs=legs,
            end_date=end_date,
            instrument_code=instrument_code,
            instrument_tag=instrument_tag,
            is_non_deliverable=is_non_deliverable,
            settlement_ccy=settlement_ccy,
            start_date=start_date,
            start_tenor=start_tenor,
            template=template,
            tenor=tenor,
            trade_date=trade_date,
        )
        super().__init__(
            definition=definition,
            fields=fields,
            pricing_parameters=pricing_parameters,
            extended_params=extended_params,
        )



from refinitiv.data.content.ipa.dates_and_calendars import is_working_day
from refinitiv.data.content.ipa.dates_and_calendars.is_working_day._is_working_day_data_provider import (
    IsWorkingDayResponseFactory,
)
from refinitiv.data.delivery._data._data_provider import ParsedData

MOCKED_IS_WORKING_DAY_DATA = ParsedData(
    {"http_status_code": 200, "http_reason": "OK"},
    {},
    [
        {
            "isWorkingDay": False,
            "isWeekEnd": False,
            "tag": "First Jan for UAH currency",
            "holidays": [
                {
                    "names": [
                        {
                            "name": "New Year's Day",
                            "calendars": ["EMU"],
                            "countries": [""],
                        }
                    ],
                    "calendars": ["EMU"],
                }
            ],
        }
    ],
)

MOCKED_DATA_ERROR = ParsedData(
    {"http_status_code": 200, "http_reason": "OK"},
    {},
    [
        {
            "error": {
                "status": "Error",
                "message": "Business Calendar: calendar not found.",
                "code": "IPA_library.ErrorBusinessCalendar_CalendarNotFound",
            },
            "tag": "First Jan for UAH currency",
        }
    ],
    ["IPA_library.ErrorBusinessCalendar_CalendarNotFound"],
    ["Business Calendar: calendar not found."],
)


def test_is_working_day():
    definition = is_working_day.Definition(
        date="2020-01-01",
        currencies=["EUR"],
        tag="First Jan for EUR currency",
        holiday_outputs=[
            is_working_day.HolidayOutputs.NAMES,
            is_working_day.HolidayOutputs.CALENDARS,
        ],
    )

    assert definition.request_item.tag == "First Jan for EUR currency"
    assert definition.request_item.date == "2020-01-01"
    assert definition.request_item.calendars is None
    assert definition.request_item.currencies == ["EUR"]
    assert definition.request_item.holiday_outputs == [
        is_working_day.HolidayOutputs.NAMES,
        is_working_day.HolidayOutputs.CALENDARS,
    ]


def test_is_working_day_create_success():
    response_factory = IsWorkingDayResponseFactory()
    response = response_factory.create_success(MOCKED_IS_WORKING_DAY_DATA)
    assert not response.data.df.empty
    assert response.data.raw == MOCKED_IS_WORKING_DAY_DATA.content_data
    assert response.data.day.holidays[0].calendars == ["EMU"]
    assert response.data.day.is_working_day is False


def test_is_working_day_create_fail():
    response_factory = IsWorkingDayResponseFactory()
    response = response_factory.create_fail(MOCKED_DATA_ERROR)
    assert response.data.raw == MOCKED_DATA_ERROR.content_data
    assert len(response.errors)



from enum import Enum
from typing import TYPE_CHECKING, Tuple

from ..._content_type import ContentType
from ..._core.session.tools import is_platform_session
from ..._tools import make_enum_arg_parser, ArgsParser, validate_bool_value

if TYPE_CHECKING:
    from ..._core.session import Session


class DataGridType(Enum):
    UDF = "udf"
    RDP = "rdp"


data_grid_types_arg_parser = make_enum_arg_parser(DataGridType)
use_field_names_in_headers_arg_parser = ArgsParser(validate_bool_value)

data_grid_type_value_by_content_type = {
    DataGridType.UDF.value: ContentType.DATA_GRID_UDF,
    DataGridType.RDP.value: ContentType.DATA_GRID_RDP,
}

content_type_by_data_grid_type = {
    ContentType.DATA_GRID_UDF: DataGridType.UDF,
    ContentType.DATA_GRID_RDP: DataGridType.RDP,
}


def get_data_grid_type(content_type: ContentType) -> DataGridType:
    data_grid_type = content_type_by_data_grid_type.get(content_type)

    if not data_grid_type:
        raise ValueError(f"There is no DataGridType for content_type:{content_type}")

    return data_grid_type


def get_content_type(session: "Session") -> ContentType:
    from ...delivery._data._data_provider_factory import get_api_config

    config = get_api_config(ContentType.DATA_GRID_RDP, session.config)
    name_platform = config.setdefault("underlying-platform", DataGridType.RDP.value)
    name_platform = data_grid_types_arg_parser.get_str(name_platform)
    content_type = data_grid_type_value_by_content_type.get(name_platform)
    return content_type


def determine_content_type_and_flag(session: "Session") -> Tuple["ContentType", bool]:
    content_type = get_content_type(session)
    changed = False

    if is_platform_session(session) and content_type == ContentType.DATA_GRID_UDF:
        content_type = ContentType.DATA_GRID_RDP
        changed = True

    return content_type, changed


def get_data_grid_type_by_session(session: "Session") -> DataGridType:
    content_type, _ = determine_content_type_and_flag(session)
    return get_data_grid_type(content_type)



# coding: utf8


from enum import Enum, unique


@unique
class Status(Enum):
    NOT_APPLICABLE = "NotApplicable"
    USER = "User"
    DATA = "Data"
    COMPUTED = "Computed"
    ERROR = "Error"
    NONE = "None"



from ._actions import Actions
from ._db_manager import DBManager, create_db_manager_by_package_name
from ._file_manager import FileManager
from .... import _configure as configure
from ....delivery.cfs._tools import (
    path_join,
    create_dirs_if_no_exists,
)
from ...._log import root_logger
from ...._tools._common import cached_property, make_counter


def get_console_logger(id_):
    logger = root_logger().getChild(f"{id_}.console")
    logger.propagate = root_logger().propagate
    return logger


def get_package_config_by_package_name(package_name):
    package_config = configure.get(f"bulk.{package_name}.package")
    if package_config is None:
        raise ValueError(f"package '{package_name}' not found in the config")
    return package_config


def create_logpath(package_name):
    package_config = get_package_config_by_package_name(package_name)
    path = package_config.get("download.path")
    create_dirs_if_no_exists(path)
    return path_join(path, "log.txt")


class _PackageManager:
    __inst_count = make_counter()

    def __init__(self, name, session=None):
        self._id = f"bulk.package_manager.{_PackageManager.__inst_count()}"
        self.name = name
        self.session = session

    @cached_property
    def actions(self) -> Actions:
        return Actions(self._id, self.logpath)

    @cached_property
    def file_manager(self) -> FileManager:
        package_config = get_package_config_by_package_name(self.name)
        package_name = package_config.get("name")
        if self.name.startswith("esg."):
            bucket = package_config.get("bucket", "bulk-ESG")
        else:
            bucket = package_config.get("bucket")
        download_path = package_config.get("download.path")
        auto_extract = package_config.get("download.auto-extract")

        if package_config.get("download.auto-retry.enabled"):
            auto_retry_count = package_config.get("download.auto-retry.count")
        else:
            auto_retry_count = 0

        file_manager = FileManager(
            actions=self.actions,
            logger=self.logger,
            package_name=package_name,
            bucket=bucket,
            path=download_path,
            auto_extract=auto_extract,
            auto_retry=auto_retry_count,
            session=self.session,
        )
        return file_manager

    @cached_property
    def db_manager(self) -> DBManager:
        return create_db_manager_by_package_name(
            self.name,
            self.actions,
            self.logger,
        )

    @cached_property
    def logpath(self):
        return create_logpath(self.name)

    @cached_property
    def logger(self):
        return get_console_logger(self._id)

    def update_files(self):
        self.file_manager.update_files()

    def reset_files(self):
        self.cleanup_files()
        self.update_files()

    def cleanup_files(self):
        self.file_manager.cleanup_files()

    def update_db(self):
        filesdata_for_update = self.file_manager.read_files_for_update()
        self.db_manager.update_db(filesdata_for_update)

    def cleanup_db(self):
        self.db_manager.cleanup_db()


def create_manager(package_name, session):
    return _PackageManager(package_name, session)



from typing import TYPE_CHECKING

from ...._object_definition import ObjectDefinition

if TYPE_CHECKING:
    from ......_types import OptStr


class RequestItem(ObjectDefinition):
    """
    Gets the Cross Currency curve definitions who can be used with provided criterias.

    Parameters
    ----------
    base_currency : str, optional

    base_index_name : str, optional

    curve_tag : str, optional
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
    quoted_currency : str, optional

    quoted_index_name : str, optional

    valuation_date : str, optional

    """

    def __init__(
        self,
        base_currency: "OptStr" = None,
        base_index_name: "OptStr" = None,
        curve_tag: "OptStr" = None,
        quoted_currency: "OptStr" = None,
        quoted_index_name: "OptStr" = None,
        valuation_date: "OptStr" = None,
    ) -> None:
        super().__init__()
        self.base_currency = base_currency
        self.base_index_name = base_index_name
        self.curve_tag = curve_tag
        self.quoted_currency = quoted_currency
        self.quoted_index_name = quoted_index_name
        self.valuation_date = valuation_date

    @property
    def base_currency(self):
        """
        :return: str
        """
        return self._get_parameter("baseCurrency")

    @base_currency.setter
    def base_currency(self, value):
        self._set_parameter("baseCurrency", value)

    @property
    def base_index_name(self):
        """
        :return: str
        """
        return self._get_parameter("baseIndexName")

    @base_index_name.setter
    def base_index_name(self, value):
        self._set_parameter("baseIndexName", value)

    @property
    def curve_tag(self):
        """
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
        :return: str
        """
        return self._get_parameter("curveTag")

    @curve_tag.setter
    def curve_tag(self, value):
        self._set_parameter("curveTag", value)

    @property
    def quoted_currency(self):
        """
        :return: str
        """
        return self._get_parameter("quotedCurrency")

    @quoted_currency.setter
    def quoted_currency(self, value):
        self._set_parameter("quotedCurrency", value)

    @property
    def quoted_index_name(self):
        """
        :return: str
        """
        return self._get_parameter("quotedIndexName")

    @quoted_index_name.setter
    def quoted_index_name(self, value):
        self._set_parameter("quotedIndexName", value)

    @property
    def valuation_date(self):
        """
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)
